# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .atm import Atm
from .gas_mix import Gas_mix
from .util.cst import PI, SIG_SB
from .util.radiation import Bnu_integral_num, rad_prop_corrk, rad_prop_xsec
from .two_stream import two_stream_toon as toon
from .two_stream import two_stream_lmdz as lmdz
from .util.spectrum import Spectrum

class Atm_2band(Atm):
    """Class based on :class:`Atm` that handles radiative trasnfer calculations
    when we want to use different radiative data to treat
    stellar absorption/scattering and the atmospheric emission.

    Only the method that change are overloaded.
    """

    def __init__(self, k_database_stellar=None, cia_database_stellar=None, **kwargs):
        """Initialization method that calls Atm_Profile().__init__() and links
        to Kdatabase and other radiative data. 
        """
        super().__init__(**kwargs)
        self.set_k_database_stellar(k_database_stellar)
        self.set_cia_database_stellar(cia_database_stellar)

    def set_k_database_stellar(self, k_database=None):
        """Change the radiative database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_k_database` for details.

        Parameters
        ----------
            k_database: :class:`Kdatabase` object
                New Kdatabase to use.
        """
        self.gas_mix_stellar.set_k_database(k_database=k_database)
        self.kdatabase_stellar=self.gas_mix_stellar.kdatabase
        self.Ng_stellar=self.gas_mix_stellar.Ng
        # to know whether we are dealing with corr-k or not and access some attributes. 
        if self.kdatabase_stellar is not None:
            self.Nw_stellar=self.kdatabase_stellar.Nw
            self.flux_top_dw_nu_stellar = np.zeros((self.Nw_stellar))

    def set_cia_database_stellar(self, cia_database=None):
        """Change the CIA database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_cia_database` for details.

        Parameters
        ----------
            cia_database: :class:`CIAdatabase` object
                New CIAdatabase to use.
        """
        self.gas_mix_stellar.set_cia_database(cia_database=cia_database)

    def set_gas(self, composition_dict, Mgas=None, compute_Mgas=True):
        """Sets the composition of the atmosphere.
        """
        for mol, vmr in composition_dict.items():
            if isinstance(vmr,(np.ndarray, list)):
                tmp_vmr=np.array(vmr)
                #geometrical average:
                composition_dict[mol]=np.sqrt(tmp_vmr[1:]*tmp_vmr[:-1])
        if self.gas_mix is None:
            self.gas_mix=Gas_mix(composition_dict)
            self.gas_mix_stellar=Gas_mix(composition_dict)
        else:
            self.gas_mix.set_composition(composition_dict)
            self.gas_mix_stellar.set_composition(composition_dict)
        if compute_Mgas: self.set_Mgas(Mgas=Mgas)

    def set_T_profile(self, tlay):
        """Reset the temperature profile without changing the pressure levels
        """
        tlay=np.array(tlay, dtype=float)
        if tlay.shape != self.logplay.shape:
            raise RuntimeError('tlay and logplay should have the same size.')
        self.tlay=tlay
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)
        self.gas_mix_stellar.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def set_adiab_profile(self, Tsurf=None, Tstrat=None):
        """Initializes the logP-T atmospheric profile with an adiabat with index R/cp=rcp

        Parameters
        ----------
            Tsurf: float
                Surface temperature.
            Tstrat: float, optional
                Temperature of the stratosphere. If None is given,
                an isothermal atmosphere with T=Tsurf is returned.
        """
        if Tstrat is None: Tstrat=Tsurf
        self.tlay=Tsurf*self.exner
        self.tlay=np.where(self.tlay<Tstrat,Tstrat,self.tlay)
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)
        self.gas_mix_stellar.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def spectral_integration_stellar(self, spectral_var):
        """Spectrally integrate an array, taking care of whether
        we are dealing with corr-k or xsec data.

        Parameters
        ----------
            spectral_var: array
                array to integrate

        Returns
        -------
            var: array
                array integrated over wavenumber (and g-space if relevant)
        """
        if self.Ng_stellar is None:
            var=np.sum(spectral_var*self.dwnedges_stellar,axis=-1)
        else:
            var=np.sum(np.sum(spectral_var*self.weights_stellar,axis=-1)*self.dwnedges_stellar,axis=-1)
        return var

    def opacity_stellar(self, rayleigh = False, **kwargs):
        """Computes the opacity of each of the radiative layers (m^2/molecule).

        Parameters
        ----------
            rayleigh: bool
                If true, the rayleigh cross section is computed in
                self.kdata_scat and added to kdata(total extinction cross section)

        See :any:`gas_mix.Gas_mix.cross_section` for details.
        """
        self.kdata_stellar = self.gas_mix_stellar.cross_section(rayleigh=rayleigh, **kwargs)
        if rayleigh: self.kdata_scat_stellar=self.gas_mix_stellar.kdata_scat
        self.Nw_stellar=self.gas_mix_stellar.Nw
        self.wns_stellar=self.gas_mix_stellar.wns
        self.wnedges_stellar=self.gas_mix_stellar.wnedges
        self.dwnedges_stellar=self.gas_mix_stellar.dwnedges

    def source_function_stellar(self, **kwargs):
        """Dummy function to have zero source in stellar channel
        """
        piBatm=np.zeros((self.Nlay,self.Nw_stellar))
        return piBatm

    def set_incoming_stellar_flux(self, flux=0., Tstar=5778., **kwargs):
        """Sets the stellar incoming flux integrated in each wavenumber
        channel.

        .. important::
            If your simulated range does not include the whole spectral range
            where the star emits, the flux seen by the model will be smaller
            than the input one. 

        Parameters
        ----------
            flux: float
                Bolometric Incoming flux (in W/m^2).
            Tstar: float
                Stellar temperature (in K) used to compute the spectral distribution
                of incoming flux using a blackbody.

        """
        self.flux_top_dw_nu_stellar = Bnu_integral_num(self.wnedges_stellar, Tstar)
        factor = flux * PI / (SIG_SB*Tstar**4 * self.dwnedges_stellar)
        self.flux_top_dw_nu_stellar = self.flux_top_dw_nu_stellar * factor


    def setup_emission_caculation_stellar(self, mu_eff=0.5, rayleigh=False,
            **kwargs):
        """Computes all necessary quantities for emission calculations
        (opacity, source, etc.)
        """
        # no need to change composition as it has been done in emission channel
        # but that won't work if you are not using both channels.
        self.opacity_stellar(rayleigh=rayleigh, **kwargs)
        self.piBatm_stellar = self.source_function_stellar()
        self.compute_layer_col_density() #done twice
        if self.Ng_stellar is None:
            self.tau_stellar, self.dtau_stellar=rad_prop_xsec(self.dcol_density_rad,
                self.kdata_stellar, mu_eff)
        else:
            self.tau_stellar, self.dtau_stellar=rad_prop_corrk(self.dcol_density_rad,
                self.kdata_stellar, mu_eff)
            self.weights_stellar=self.kdatabase_stellar.weights

    def reflexion_spectrum_2stream(self, integral=True, mu0_stellar=0.5,
            method='toon', dtau_min=1.e-10, flux_at_level=False, rayleigh=False,
            flux_top_dw=None, **kwargs):
        """Returns the reflexion flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation_stellar(mu_eff=1., rayleigh=rayleigh, integral=integral,
            flux_top_dw=flux_top_dw, **kwargs)
        # mu_eff=1. because the mu effect is taken into account in solve_2stream_nu
                 # we must compute the vertical optical depth here.
        self.dtau_stellar=np.where(self.dtau_stellar<dtau_min,dtau_min,self.dtau_stellar)

        module_to_use=globals()[method]
        # globals()[method] converts the method string into a module name
        #  if the module has been loaded
        if self.Ng_stellar is None:
            solve_2stream_nu=module_to_use.solve_2stream_nu_xsec
        else:
            solve_2stream_nu=module_to_use.solve_2stream_nu_corrk

        if rayleigh:
            if self.Ng_stellar is None:
                self.single_scat_albedo_stellar = self.kdata_scat_stellar / self.kdata_stellar
            else:
                self.single_scat_albedo_stellar = self.kdata_scat_stellar[:,:,None] / self.kdata_stellar
        else:
            self.single_scat_albedo_stellar = np.zeros_like(self.dtau)
        self.single_scat_albedo_stellar=np.core.umath.minimum(self.single_scat_albedo_stellar,0.9999999999999)
        self.asym_param_stellar = np.zeros_like(self.dtau_stellar)

        if flux_top_dw is not None:
            self.set_incoming_stellar_flux(flux=flux_top_dw, **kwargs)

        self.flux_up_nu_stellar, self.flux_down_nu_stellar, self.flux_net_nu_stellar = \
            solve_2stream_nu(self.piBatm_stellar, self.dtau_stellar,
                self.single_scat_albedo_stellar, self.asym_param_stellar,
                self.flux_top_dw_nu_stellar, mu0 = mu0_stellar, flux_at_level=flux_at_level)

        if self.Ng_stellar is None:
            return Spectrum(self.flux_up_nu_stellar[0],self.wns_stellar,self.wnedges_stellar)
        else:
            return Spectrum(np.sum(self.flux_up_nu_stellar[0]*self.weights_stellar,axis=1),
                        self.wns_stellar,self.wnedges_stellar)

    def flux_divergence_stellar(self, per_unit_mass = True, **kwargs):
        """Computes the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface.
            net: array
                Net fluxes at level surfaces
        """
        if self.flux_net_nu_stellar is None:
            raise RuntimeError('should have ran reflexion_spectrum_2stream.')
        net=self.spectral_integration_stellar(self.flux_net_nu_stellar)
        H=-np.copy(net)
        #print(H)
        H[:-1]-=H[1:]
        if per_unit_mass: H*=self.inv_dmass
        return H, net

    def heating_rate(self, compute_kernel=False, dTmax_use_kernel=None,
            flux_top_dw=None, **kwargs):
        if (not compute_kernel) and (dTmax_use_kernel is not None):
            dT=self.tlay-self.tlay_kernel
            if np.amax(np.abs(dT)) < dTmax_use_kernel:
                try:
                    H = self.H_kernel + np.dot(dT,self.kernel)
                except:
                    raise RuntimeError("Kernel has not been precomputed")
                net = np.zeros_like(H)
                return H, net
        _ = self.emission_spectrum_2stream(flux_at_level=True, integral=True,
                compute_kernel=compute_kernel, flux_top_dw=None, **kwargs)
        H, net = self.flux_divergence(compute_kernel=compute_kernel, **kwargs)
        _ = self.reflexion_spectrum_2stream(flux_at_level=True, integral=True,
                flux_top_dw=flux_top_dw, **kwargs)
        H_stellar, net_stellar = self.flux_divergence_stellar(**kwargs)
        #print(H, H_stellar)
        if compute_kernel:
            self.H_kernel = H + H_stellar
            self.tau_rad = 1./np.amax(np.abs(self.kernel.diagonal()))
        return H+H_stellar, net+net_stellar


    def bolometric_fluxes_stellar(self, per_unit_mass = True):
        """Computes the bolometric fluxes at levels and the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            up: array
                Upward fluxes at level surfaces
            dw: array
                Downward fluxes at level surfaces
            net: array
                Net fluxes at level surfaces
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface.
        """
        H, net = self.flux_divergence_stellar(per_unit_mass = per_unit_mass)
        up=self.spectral_integration_stellar(self.flux_up_nu_stellar)
        dw=self.spectral_integration_stellar(self.flux_down_nu_stellar)
        return up, dw, net, H

    def __repr__(self):
        """Method to output header
        """
        output=super().__repr__()
        output+="""
    k_database_ste  :
        {kdatab}
    cia_database_ste:
        {cdatab}""".format(kdatab=self.kdatabase, cdatab=self.gas_mix.cia_database)
        if self.gas_mix._wn_range is not None:
            output+='    wn range        : '+ self.gas_mix._wn_range +'\n'

        return output
        



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Class for Rayleigh opacties. 
"""
import numpy as np
from .util.cst import PI, KBOLTZ
from .util.singleton import Singleton

class Rayleigh(Singleton):
    """Class to compute Rayleigh opacities
    """

    def init(self, *args, **kwds):
        """Initializes various parameters for Rayleigh computations
        """
        N_std=1.e5/(KBOLTZ*273.15)
        self._mult_factor=24.*PI**3/(N_std)**2
        self._mult_factor=self._mult_factor*100.**4
        # last 100.**4 is because we are using wavenumbers in cm^-1
        # instead of wavelength in m (see eq 12 of Caldas 2019)

    def sigma(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.
        This one is faster than sigma_array, but can be used only
        when vmr values are constants.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of arrays
                Keys are molecule names. Values are the volume mixing ratios.
                For speedup, only the first value will be used because we assume
                that the vmr arrays are constant

        Returns
        -------
            array of shape (wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        res=np.zeros(wns.size)
        wn2 = wns*wns
        wn4 = wn2*wn2
        for mol, x in vmr.items():
            to_add, tmp = self.sigma_mol(mol, wn2, wn4)
            if to_add: res+=x[0]*tmp

        return res

    def sigma_array(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of arrays
                Keys are molecule names. Values are arrays the volume mixing ratios

        Returns
        -------
            array of shape (vmr.values.size, wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        first_mol=True
        wn2 = wns*wns
        wn4 = wn2*wn2
        for mol, x in vmr.items():
            x_array=np.array(x)
            if first_mol:
                res=np.zeros((x_array.size,wns.size))
                first_mol=False
            to_add, tmp = self.sigma_mol(mol, wn2, wn4)
            if to_add: res+=x_array[:,None]*tmp

        return res

    def sigma_mol(self, mol, wn2, wn4):
        """Intermediary function to compute rayleigh for each molecule.

        Parameters
        ----------
            mol: str
                Molecule name.
            wn2, wn4: arrays
                Array of the wavenumber (in cm^-1) to the 2nd and 4th power.
                (To avoid recomputing it each time).
        
        Returns
        -------
            to_add: bool
                Says whether the molecule has been found
                and the contribution needs to be added.
            tmp: array of size self.wns or None
                The cross section for the molecule as a function of wns.
                None if the molecule has not
                been found.

        """
        to_add=True
        tmp=None
        if mol == 'H2':
            #tmp=((8.14E-13)*(wave**(-4.))* \
            #    (1+(1.572E6)*(wave**(-2.))+(1.981E12)*(wave**(-4.))))*1E-4
            tmp=(8.14e-49+1.28e-58*wn2+1.61e-67*wn4)*wn4
        elif mol == 'He':
            #tmp=((5.484E-14)*(wave**(-4.))*(1+(2.44E5)*(wave**(-2.))))*1E-4
            tmp=(5.484e-50+1.338e-60*wn2)*wn4
        elif mol=='N2':
            tmp=self._mult_factor * wn4 * (1.034 + 3.17e-12*wn2) * \
                (6.4982e-5 + 3.0743305e-10/(1.44e10-wn2))**2 * 4./9.
                # 4./9. is approximately ((n+1)/(n**2+2))**2
                # from eq 12 of caldas et al. 2019
        elif mol=='O2':
            tmp=self._mult_factor * wn4 * (1.096 + 1.385e-11*wn2 + 1.448e-20*wn4) * \
                (2.1351e-4 + 0.218567e-10/(0.409e10-wn2))**2 * 4./9.
        elif mol=='CO2':
            tmp=self._mult_factor * wn4 * (1.1364 + 2.53e-11*wn2) * \
                (1.1427e-2 * (5.799e3/(1.6617e10-wn2) + 1.2005e2/(7.960e9-wn2) \
                    + 5.3334/(5.630e9-wn2)))**2 * 4./9.
        else:
            to_add=False
        return to_add, tmp



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

This module contain classes to handle atmospheres
and their radiative properties.
This alows us to compute the transmission and emission spectra
of those atmospheres.

The nomenclature for layers, levels, etc, is as follows (example with 4 layers; Nlay=4):

::

    -------------------  Model top or first level (plev[0])
    - - - - - - - - - -  First atmopheric layer (play[0]=plev[0], tlay[0], xlay[0])

    -------------------  plev[1]

    - - - - - - - - - -  play[1], tlay[1], xlay[1]

    -------------------  plev[2]

    - - - - - - - - - - play[2], tlay[2], xlay[2]

    -------------------  plev[3]

    - - - - - - - - - -  bottom layer (play[Nlay-1]=psurf, tlay[3], xlay[3])
    ------------------- Surface (plev[Nlev-1=Nlay]=psurf)
    ///////////////////

.. image:: /images/atm_schema.png


Tempratures (`tlay`) and volume mixing ratios (`xlay`) are provided at the
*mid layer* point (`play`) (Note that this does not mean
that this point is the middle of the layer, in general, it is not).

If pressure levels are not specified by the user with `logplevel`,
they are at the mid point between
the pressure of the layers directly above and below. The pressure of the
top and bottom levels are counfounded with the pressure of the
top and bottom mid layer points.

For radiative calculations, the source function (temperature) needs to be known at
the boudaries of the *radiative* layers but the opacity needs to be known
inside the *radiative* layer.
For this reason, there are `Nlay-1` radiative layers and they are offset
with respect to atmospheric layers.
Opacities are computed at the center of those radiative layers, i.e. at the pressure levels.
The temperature is interpolated at these levels with an arithmetic average.
Volume mixing ratios are interpolated using a geometric average. 

"""
import numpy as np
import astropy.units as u
from numba.typed import List
from .gas_mix import Gas_mix
from .aerosols import Aerosols
from .util.cst import N_A, PI, RGP, KBOLTZ, SIG_SB
from .util.radiation import Bnu_integral_num, Bnu, rad_prop_corrk, rad_prop_xsec,\
    Bnu_integral_array, path_integral_corrk, path_integral_xsec
from .two_stream import two_stream_toon as toon
from .two_stream import two_stream_lmdz as lmdz
from .util.interp import gauss_legendre
from .util.spectrum import Spectrum


class Atm_profile(object):
    """A class defining an atmospheric PT profile with some global data
    (gravity, etc.)

    The derived class :class:`~exo_k.atm.Atm` handles
    radiative transfer calculations.

    """
    
    def __init__(self, composition={}, psurf=None, ptop=None, logplay=None, tlay=None,
            Tsurf=None, Tstrat=None, grav=None, Rp=None, Mgas=None, rcp=0.28, Nlay=20,
            logplev=None, aerosols=None,
            ## old parameters that should be None. THey are left here to catch
            ## exceptions and warn the user that their use is obsolete
            Nlev=None, tlev=None,
            **kwargs):
        """Initializes atmospheric profiles

        Parameters
        ----------
            composition: dict
                Keys are molecule names and values the vmr.
                Vmr can be arrays of size Nlev-1 (i.e. the number of layers).
            grav: float
                Planet surface gravity (gravity constant with altitude for now).
            Rp: float or Astropy.unit quantity
                Planet radius. If float, meters are assumed.
            rcp: float
                Adiabatic lapse rate for the gas (R/cp)
            Mgas: float, optional
                Molar mass of the gas (kg/mol). If given, overrides the molar mass computed
                from composition.
        
        There are two ways to define the profile.
        You can define:

        * Nlay: int
          Number of layers
        * psurf, Tsurf: float
          Surface pressure (Pa) and temperature 
        * ptop: float
          Pressure at the top of the model (Pa) 
        * Tstrat: float
          Stratospheric temperature        

        This way you will have an adiabatic atmosphere with Tsurf at the ground that
        becomes isothermal wherever T=<Tstrat.
        You can also specify:

        * logplay or play: array
        * tlay: array (same size)
          These will become the pressures (Pa; the log10 if you give
          logplay) and temperatures of the layers.
          This will be used to define the surface and top pressures.
          Nlay becomes the size of the arrays. 

        .. warning::
            Layers are counted from the top down (increasing pressure order).
            All methods follow the same convention.
        """
        if (Nlev is not None) or (tlev is not None):
            print("""
                since version 1.1.0, Nlev, tlev, and plev have been renamed
                Nlay, tlay, and logplay for consistency with other codes.
                Just change the name of the variables in the method call
                and you should be just fine!
                """)
            raise RuntimeError('Unknown keyword argument in __init__')
        self.gas_mix = None
        self.set_gas(composition, compute_Mgas=False)
        self.set_aerosols(aerosols)
        self.rcp = rcp
        self.logplev = None
        if logplay is None:
            self.Nlay = Nlay
            self.Nlev = Nlay+1
            self.logplay = np.linspace(np.log10(ptop),np.log10(psurf),num=self.Nlay)
            self.compute_pressure_levels()
            self.set_adiab_profile(Tsurf=Tsurf, Tstrat=Tstrat)
        else:
            self.set_logPT_profile(logplay, tlay, logplev=logplev)
        self.set_Rp(Rp)        
        self.set_grav(grav)
        self.set_Mgas(Mgas=Mgas)

    def set_logPT_profile(self, logplay, tlay, logplev=None):
        """Set the logP-T profile of the atmosphere with a new one

        Parameters
        ----------
            logplay: array
                Log pressure (in Pa) of the layer
            tlay: array (same size)
                temperature of the layers.
        
        Other Parameters
        ----------------
            logplev: array (size Nlay+1)
                If provided, allows the user to choose the location
                of the level surfaces separating the layers.
        """
        self.logplay=np.array(logplay, dtype=float)
        self.Nlay=self.logplay.size
        self.Nlev=self.Nlay+1
        if logplev is not None:
            if logplev.size == self.Nlev:
                self.logplev=np.array(logplev, dtype=float)
            else:
                raise RuntimeError('logplev does not have the size Nlay+1')
        self.compute_pressure_levels()
        self.set_T_profile(tlay)

    def set_T_profile(self, tlay):
        """Reset the temperature profile without changing the pressure levels
        """
        tlay=np.array(tlay, dtype=float)
        if tlay.shape != self.logplay.shape:
            raise RuntimeError('tlay and logplay should have the same size.')
        self.tlay=tlay
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def compute_pressure_levels(self):
        """Computes various pressure related quantities
        """
        if self.logplay[0] >= self.logplay[-1]:
            print("""
            Atmospheres are modelled from the top down.
            All arrays should be ordered accordingly
            (first values correspond to top of atmosphere)""")
            raise RuntimeError('Pressure grid is in decreasing order!')
        self.play=np.power(10., self.logplay)
        if self.logplev is None:
        # case where the levels are halfway between layer centers
            self.plev=np.zeros(self.Nlev)
            self.plev[1:-1]=(self.play[:-1]+self.play[1:])*0.5
            # we choose to use mid point so that there is equal mass in the bottom half
            # of any top layer and the top half of the layer below. 
            self.plev[0]=self.play[0]
            self.plev[-1]=self.play[-1]
            ## WARNING: Top and bottom pressure levels are set equal to the
            #  pressure in the top and bottom layers. If you change that,
            #  some assumptions here and there in the code may break down!!!
            self.logplev=np.log10(self.plev)

        # case where the levels are halfway between layer centers in LOG10
            #self.logplev=np.zeros(self.Nlev)
            #self.logplev[1:-1]=(self.logplay[:-1]+self.logplay[1:])*0.5
            #self.logplev[0]=self.logplay[0]
            #self.logplev[-1]=self.logplay[-1]
            #self.plev=np.power(10.,self.logplev)
        else:
            self.plev=np.power(10., self.logplev)

        self.logp_opac=self.logplev[1:-1]
        self.psurf=self.plev[-1]
        self.dp_lay=np.diff(self.plev) ### probably redundant with dmass
        self.exner=(self.play/self.psurf)**self.rcp

    def set_adiab_profile(self, Tsurf=None, Tstrat=None):
        """Initializes the logP-T atmospheric profile with an adiabat with index R/cp=rcp

        Parameters
        ----------
            Tsurf: float
                Surface temperature.
            Tstrat: float, optional
                Temperature of the stratosphere. If None is given,
                an isothermal atmosphere with T=Tsurf is returned.
        """
        if Tstrat is None: Tstrat=Tsurf
        self.tlay=Tsurf*self.exner
        self.tlay=np.where(self.tlay<Tstrat,Tstrat,self.tlay)
        self.t_opac=(self.tlay[:-1]+self.tlay[1:])*0.5
        self.gas_mix.set_logPT(logp_array=self.logp_opac, t_array=self.t_opac)

    def set_grav(self, grav=None):
        """Sets the surface gravity of the planet

        Parameters
        ----------
            grav: float
                surface gravity (m/s^2)
        """
        if grav is None: raise RuntimeError('A planet needs a gravity!')
        self.grav=grav
        self.dmass=self.dp_lay/self.grav
        self.inv_dmass=1./self.dmass
    
    def set_gas(self, composition_dict, Mgas=None, compute_Mgas=True):
        """Sets the composition of the atmosphere.

        The composition_dict gives the composition in the layers, but we will
        need the composition in the radiative layers, so the interpolation is
        done here. For the moment we do a geometrical average.

        .. important::
            For the first initialization, compute_Mgas must be False
            because we need Gas_mix to be initialized before we know the number of layers
            in the atmosphere, but the number of layers is needed by set_Mgas!
            So set_Mgas needs to be called at the end of the initialization.

        Parameters
        ----------
            composition_dict: dictionary
                Keys are molecule names, and values are volume mixing ratios.
                A 'background' value means that the gas will be used to fill up to vmr=1
                If they do not add up to 1 and there is no background gas_mix,
                the rest of the gas_mix is considered transparent.
            compute_Mgas: bool
                If False, the molar mass of the gas is not updated. 
        """
        for mol, vmr in composition_dict.items():
            if isinstance(vmr,(np.ndarray, list)):
                tmp_vmr=np.array(vmr)
                #geometrical average:
                composition_dict[mol]=np.sqrt(tmp_vmr[1:]*tmp_vmr[:-1])
        if self.gas_mix is None:
            self.gas_mix=Gas_mix(composition_dict)
        else:
            self.gas_mix.set_composition(composition_dict)
        if compute_Mgas: self.set_Mgas(Mgas=Mgas)

    def set_Mgas(self, Mgas=None):
        """Sets the mean molar mass of the atmosphere.

        Parameters
        ----------
            Mgas_rad: float or array of size Nlay-1
                Mean molar mass in the radiative layers (kg/mol).
                If None is given, the mmm is computed from the composition.
        """
        if Mgas is not None:
            self.Mgas_rad=Mgas
        else:
            self.Mgas_rad=self.gas_mix.molar_mass()
        if not isinstance(self.Mgas_rad, np.ndarray):
            self.Mgas_rad=self.Mgas_rad*np.ones(self.Nlay-1, dtype=np.float)

    def set_rcp(self,rcp):
        """Sets the adiabatic index of the atmosphere

        Parameters
        ----------
            rcp: float
                R/c_p
        """
        self.rcp=rcp

    def set_aerosols(self, aerosols):
        """Sets the aerosols dictionary
        """
        self.aerosols = Aerosols(aerosols)

    def set_Rp(self, Rp):
        """Sets the radius of the planet

        Parameters
        ----------
            Rp: float
                radius of the planet (m)
        """
        if Rp is None:
            self.Rp = None
            return
        if isinstance(Rp,u.quantity.Quantity):
            self.Rp=Rp.to(u.m).value
        else:
            self.Rp=Rp

    def set_Rstar(self, Rstar):
        """Sets the radius of the star

        Parameters
        ----------
            Rstar: float
                radius of the star (m)
        """
        if Rstar is None:
            self.Rstar = None
            return
        if isinstance(Rstar,u.quantity.Quantity):
            self.Rstar=Rstar.to(u.m).value
        else:
            self.Rstar=Rstar

    def compute_density(self):
        """Computes the number density (m^-3) profile of the atmosphere
        in the radiative layers
        """
        self.density=np.power(10., self.logp_opac)/(KBOLTZ*self.t_opac)

    def compute_layer_col_density(self):
        """Computes the column number density (molecules/m^2) per
        radiative layer of the atmosphere.

        There are Nlay-1 radiative layers as they go from the midle of a layer to the next.
        """
        factor=N_A/(self.grav * self.Mgas_rad)
        self.dcol_density_rad = np.diff(self.play)*factor[:]

        if self.Rp is not None: #includes the altitude effect if radius is known
            self.compute_altitudes()
            self.dcol_density_rad*=(1.+self.zlev[1:-1]/self.Rp)**2

    def compute_altitudes(self):
        """Compute altitudes of the level surfaces (zlev) and mid layers (zlay).
        """
        Mgas = np.empty(self.Nlay, dtype=np.float)
        Mgas[1:-1] = 0.5*(self.Mgas_rad[:-1]+self.Mgas_rad[1:])
        Mgas[0] = self.Mgas_rad[0]
        Mgas[-1] = self.Mgas_rad[-1]
        H = RGP*self.tlay/(self.grav*Mgas)
        dlnP = np.diff(self.logplev)*np.log(10.)
        self.zlev = np.zeros_like(self.logplev)
        if self.Rp is None:
            self.dz = H*dlnP
            self.zlev[:-1] = np.cumsum(self.dz[::-1])[::-1]
        else:
            for i in range(H.size)[::-1]:
                z1 = self.zlev[i+1]
                H1 = H[i]
                dlnp = dlnP[i]
                self.zlev[i] = z1+( (H1 * (self.Rp + z1)**2 * dlnp) \
                    / (self.Rp**2 + H1 * self.Rp * dlnp + H1 * z1 * dlnp) )
        self.zlay = 0.5*(self.zlev[1:]+self.zlev[:-1])
        self.zlay[-1] = 0.
        self.zlay[0] = self.zlev[0]
        ## assumes layer centers at the middle of the two levels
        ## which is not completely consistent with play, but should be
        ## a minor error.
        
    def compute_area(self):
        """Computes the area of the annulus covered by each
        radiative layer (from a mid layer to the next) in a transit setup. 
        """
        self.area=PI*(self.Rp+self.zlay[:-1])**2
        self.area[:-1]-=self.area[1:]
        self.area[-1]-=PI*self.Rp**2

    def compute_tangent_path(self):
        """Computes a triangular array of the tangent path length (in m) spent in each
        radiative layer.
        
        self.tangent_path[ilay][jlay] is the length that the ray that is tangent to the ilay 
        radiative layer spends in the jlay>=ilay layer
        (accounting for a factor of 2 due to symmetry)
        """
        if self.Rp is None: raise RuntimeError('Planetary radius should be set')
        self.compute_altitudes()
        self.tangent_path=List()
        # List() is a new numba.typed list to comply with new numba evolution after v0.50
        for ilay in range(self.Nlay-1): #layers counted from the top
            z0square=(self.Rp+self.zlev[ilay+1])**2
            dl=np.sqrt((self.Rp+self.zlay[:ilay+1])**2-z0square)
            dl[:-1]-=dl[1:]
            self.tangent_path.append(2.*dl)

    def __repr__(self):
        """Method to output header
        """
        output="""
    gravity (m/s^2) : {grav}
    Planet Radius(m): {rad}
    Ptop (Pa)       : {ptop}
    Psurf (Pa)      : {psurf}
    Tsurf (K)       : {tsurf}
    composition     :
        {comp}""".format(grav=self.grav, rad=self.Rp, comp=self.gas_mix,
            ptop=self.plev[0], psurf=self.psurf, tsurf=self.tlay[-1])
        return output



class Atm(Atm_profile):
    """Class based on Atm_profile that handles radiative trasnfer calculations.

    Radiative data are accessed through the :any:`gas_mix.Gas_mix` class.
    """

    def __init__(self, k_database=None, cia_database=None, a_database=None,
        wn_range=None, wl_range=None, internal_flux=0., **kwargs):
        """Initialization method that calls Atm_Profile().__init__() and links
        to Kdatabase and other radiative data. 
        """
        super().__init__(**kwargs)
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)
        self.set_a_database(a_database)
        self.set_spectral_range(wn_range=wn_range, wl_range=wl_range)
        self.set_internal_flux(internal_flux)
        self.flux_net_nu=None
        self.kernel=None

    def set_k_database(self, k_database=None):
        """Change the radiative database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_k_database` for details.

        Parameters
        ----------
            k_database: :class:`Kdatabase` object
                New Kdatabase to use.
        """
        self.gas_mix.set_k_database(k_database=k_database)
        self.kdatabase=self.gas_mix.kdatabase
        self.Ng=self.gas_mix.Ng
        # to know whether we are dealing with corr-k or not and access some attributes. 
        if self.kdatabase is not None:
            self.Nw=self.kdatabase.Nw
            self.flux_top_dw_nu = np.zeros((self.Nw))

    def set_cia_database(self, cia_database=None):
        """Change the CIA database used by the
        :class:`Gas_mix` object handling opacities inside
        :class:`Atm`.

        See :any:`gas_mix.Gas_mix.set_cia_database` for details.

        Parameters
        ----------
            cia_database: :class:`CIAdatabase` object
                New CIAdatabase to use.
        """
        self.gas_mix.set_cia_database(cia_database=cia_database)

    def set_a_database(self, a_database=None):
        """Change the Aerosol database used by the
        :class:`Aerosols` object handling aerosol optical properties.

        See :any:`aerosols.Aerosols.set_a_database` for details.

        Parameters
        ----------
            a_database: :class:`Adatabase` object
                New Adatabase to use.
        """    
        self.aerosols.set_a_database(a_database=a_database)

    def set_spectral_range(self, wn_range=None, wl_range=None):
        """Sets the spectral range in which computations will be done by specifying
        either the wavenumber (in cm^-1) or the wavelength (in micron) range.

        See :any:`gas_mix.Gas_mix.set_spectral_range` for details.
        """
        self.gas_mix.set_spectral_range(wn_range=wn_range, wl_range=wl_range)

    def set_incoming_stellar_flux(self, flux=0., Tstar=5778., **kwargs):
        """Sets the stellar incoming flux integrated in each wavenumber
        channel.

        .. important::
            If your simulated range does not include the whole spectral range
            where the star emits, the flux seen by the model will be smaller
            than the input one. 

        Parameters
        ----------
            flux: float
                Bolometric Incoming flux (in W/m^2).
            Tstar: float
                Stellar temperature (in K) used to compute the spectral distribution
                of incoming flux using a blackbody.

        """
        self.flux_top_dw_nu = Bnu_integral_num(self.wnedges, Tstar)
        factor = flux * PI / (SIG_SB*Tstar**4 * self.dwnedges)
        self.flux_top_dw_nu = self.flux_top_dw_nu * factor
    
    def set_internal_flux(self, internal_flux):
        """Sets internal flux from the subsurface in W/m^2
        """
        self.internal_flux = internal_flux

    def spectral_integration(self, spectral_var):
        """Spectrally integrate an array, taking care of whether
        we are dealing with corr-k or xsec data.

        Parameters
        ----------
            spectral_var: array
                array to integrate

        Returns
        -------
            var: array
                array integrated over wavenumber (and g-space if relevant)
        """
        if self.Ng is None:
            var=np.sum(spectral_var*self.dwnedges,axis=-1)
        else:
            var=np.sum(np.sum(spectral_var*self.weights,axis=-1)*self.dwnedges,axis=-1)
        return var

    def opacity(self, rayleigh = False, compute_all_opt_prop = False, **kwargs):
        """Computes the opacity of each of the radiative layers (m^2/molecule).

        Parameters
        ----------
            rayleigh: bool
                If true, the rayleigh cross section is computed in
                self.kdata_scat and added to kdata(total extinction cross section)

        See :any:`gas_mix.Gas_mix.cross_section` for details.
        """
        self.kdata = self.gas_mix.cross_section(rayleigh=rayleigh, **kwargs)
        shape = self.kdata.shape
        if self.aerosols.adatabase is not None:
            [kdata_aer, k_scat_aer, g_aer] = self.aerosols.optical_properties(
                    compute_all_opt_prop=compute_all_opt_prop, **kwargs)
            if self.Ng is None:
                self.kdata += kdata_aer
            else:
                self.kdata += kdata_aer[:,:,None]
        if compute_all_opt_prop:
            if rayleigh:
                kdata_scat_tot = self.gas_mix.kdata_scat
            else:
                kdata_scat_tot = np.zeros(shape[0:2], dtype=np.float)
            if self.aerosols.adatabase is not None:
                kdata_scat_tot += k_scat_aer
                self.asym_param = k_scat_aer * g_aer / kdata_scat_tot
            else:
                self.asym_param = np.zeros(shape[0:2], dtype=np.float)

            if self.Ng is None:
                self.single_scat_albedo = kdata_scat_tot / self.kdata
            else:
                self.single_scat_albedo = kdata_scat_tot[:,:,None] / self.kdata
                self.asym_param = self.asym_param[:,:,None] * np.ones(self.Ng)
                #JL21 the line above could be removed as asym param does not seem to change
                # with g point, but then the dimensions should be change in 2stream routines.
            self.single_scat_albedo=np.core.umath.minimum(self.single_scat_albedo,0.9999999999999)

        self.Nw=self.gas_mix.Nw
        self.wns=self.gas_mix.wns
        self.wnedges=self.gas_mix.wnedges
        self.dwnedges=self.gas_mix.dwnedges

    def source_function(self, integral=True, source=True):
        """Compute the blackbody source function (Pi*Bnu) for each layer of the atmosphere.

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
            source: boolean, optional
                If False, the source function is put to 0 (for solar absorption calculations)
        """
        if source:
            if integral:
                #JL2020 What is below is slightly faster for very large resolutions
                #piBatm=np.empty((self.Nlay,self.Nw))
                #for ii in range(self.Nlay):
                #    piBatm[ii]=PI*Bnu_integral_num(self.wnedges,self.tlay[ii])/dw
                #JL2020 What is below is much faster for moderate to low resolutions
                piBatm=PI*Bnu_integral_array(self.wnedges,self.tlay,self.Nw,self.Nlay) \
                    /self.dwnedges
            else:
                piBatm=PI*Bnu(self.wns[None,:],self.tlay[:,None])
        else:
            piBatm=np.zeros((self.Nlay,self.Nw))
        return piBatm

    def setup_emission_caculation(self, mu_eff=0.5, rayleigh=False, integral=True,
            source=True, gas_vmr=None, **kwargs):
        """Computes all necessary quantities for emission calculations
        (opacity, source, etc.)
        """
        if gas_vmr is not None: self.set_gas(gas_vmr)
        self.opacity(rayleigh=rayleigh, **kwargs)
        self.piBatm = self.source_function(integral=integral, source=source)
        self.compute_layer_col_density()
        if self.Ng is None:
            self.tau, self.dtau=rad_prop_xsec(self.dcol_density_rad,
                self.kdata, mu_eff)
        else:
            self.tau, self.dtau=rad_prop_corrk(self.dcol_density_rad,
                self.kdata, mu_eff)
            self.weights=self.kdatabase.weights

    def emission_spectrum(self, integral=True, mu0=0.5, mu_quad_order=None,
            dtau_min=1.e-13, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            mu_quad_order: int
                If an integer is given, the emission intensity is computed
                for a number of angles and integrated following a gauss legendre
                quadrature rule of order `mu_quad_order`.
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        if mu_quad_order is not None:
            # if we want quadrature, use the more general method.
            return self.emission_spectrum_quad(integral=integral,
                mu_quad_order=mu_quad_order, dtau_min=dtau_min, **kwargs)

        try:
            self.setup_emission_caculation(mu_eff=mu0, rayleigh=False, integral=integral, **kwargs)
        except TypeError:
            raise RuntimeError("""
            Cannot use rayleigh option with emission_spectrum.
            If you meant to include scattering, you should use emission_spectrum_2stream.
            """)
        # self.tau and self.dtau include the 1/mu0 factor.
        expdtau=np.exp(-self.dtau)
        expdtauminone=np.where(self.dtau < dtau_min, -self.dtau, expdtau-1.)
        # careful: due to numerical limitations, 
        # the limited development of Exp(-dtau)-1 needs to be used for small values of dtau
        exptau=np.exp(-self.tau)
        if self.Ng is None:
            timesBatmTop=(1.+expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom=(-expdtau-expdtauminone/self.dtau)*exptau[:-1]
            timesBatmBottom[-1]+=exptau[-1]
        else:
            timesBatmTop=np.sum((1.+expdtauminone/self.dtau)*exptau[:-1]*self.weights,axis=-1)
            timesBatmBottom=np.sum((-expdtau-expdtauminone/self.dtau)*exptau[:-1] \
                *self.weights,axis=-1)
            timesBatmBottom[-1]+=np.sum(exptau[-1]*self.weights,axis=-1)
        IpTop=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0)

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_quad(self, integral=True, mu_quad_order=3, dtau_min=1.e-13, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)
        using gauss legendre qudrature of order `mu_quad_order`

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.
                  
        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh=False, integral=integral, **kwargs)
        # angle effect dealt with later

        IpTop=np.zeros(self.kdata.shape[1])
        mu_w, mu_a, _ = gauss_legendre(mu_quad_order)
        mu_w = mu_w * mu_a * 2.# takes care of the mu factor in last integral => int(mu I d mu)
                               # Factor 2 takes care of the fact that the source function is pi*Batm
                               # but we want 2*Pi*Batm
        for ii, mu0 in enumerate(mu_a):
            tau=self.tau/mu0
            dtau=self.dtau/mu0
            expdtau=np.exp(-dtau)
            expdtauminone=np.where(dtau<dtau_min,-dtau,expdtau-1.)
            exptau=np.exp(-tau)
            if self.Ng is None:
                timesBatmTop=(1.+expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom=(-expdtau-expdtauminone/dtau)*exptau[:-1]
                timesBatmBottom[-1]+=exptau[-1]
            else:
                timesBatmTop=np.sum((1.+expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom=np.sum((-expdtau-expdtauminone/dtau)*exptau[:-1] \
                    *self.weights,axis=-1)
                timesBatmBottom[-1]+=np.sum(exptau[-1]*self.weights,axis=-1)
            IpTop+=np.sum(self.piBatm[:-1]*timesBatmTop+self.piBatm[1:]*timesBatmBottom,axis=0) \
                *mu_w[ii]

        return Spectrum(IpTop,self.wns,self.wnedges)

    def emission_spectrum_2stream(self, integral=True, mu0=0.5,
            method='toon', dtau_min=1.e-10, flux_at_level=False, rayleigh=False,
            flux_top_dw=None, source=True, compute_kernel=False, **kwargs):
        """Returns the emission flux at the top of the atmosphere (in W/m^2/cm^-1)

        Parameters
        ----------
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 

        Other Parameters
        ----------------
            mu0: float
                Cosine of the quadrature angle use to compute output flux
            dtau_min: float
                If the optical depth in a layer is smaller than dtau_min,
                dtau_min is used in that layer instead. Important as too
                transparent layers can cause important numerical rounding errors.

        Returns
        -------
            Spectrum object 
                A spectrum with the Spectral flux at the top of the atmosphere (in W/m^2/cm^-1)
        """
        self.setup_emission_caculation(mu_eff=1., rayleigh=rayleigh, integral=integral,
            source=source, flux_top_dw=flux_top_dw, compute_all_opt_prop=True, **kwargs)
        # mu_eff=1. because the mu effect is taken into account in solve_2stream_nu
                 # we must compute the vertical optical depth here.
        # JL21: shouldn't we remove flux_top_dw, it seems not to be used.
        self.dtau=np.where(self.dtau<dtau_min,dtau_min,self.dtau)

        module_to_use=globals()[method]
        # globals()[method] converts the method string into a module name
        #  if the module has been loaded
        if self.Ng is None:
            solve_2stream_nu=module_to_use.solve_2stream_nu_xsec
        else:
            solve_2stream_nu=module_to_use.solve_2stream_nu_corrk

        if flux_top_dw is not None:
            self.set_incoming_stellar_flux(flux=flux_top_dw, **kwargs)

        self.flux_up_nu, self.flux_down_nu, self.flux_net_nu = \
            solve_2stream_nu(self.piBatm, self.dtau, self.single_scat_albedo, self.asym_param,
                self.flux_top_dw_nu, mu0 = mu0, flux_at_level=flux_at_level)
        if compute_kernel: self.compute_kernel(solve_2stream_nu, mu0=mu0, flux_at_level=flux_at_level,
                    per_unit_mass=True, integral=True, **kwargs)

        if self.Ng is None:
            return Spectrum(self.flux_up_nu[0],self.wns,self.wnedges)
        else:
            return Spectrum(np.sum(self.flux_up_nu[0]*self.weights,axis=1),self.wns,self.wnedges)

    def compute_kernel(self, solve_2stream_nu, epsilon=0.01, flux_at_level=False, mu0 = 0.5,
            per_unit_mass=True, integral=True, **kwargs):
        """Compute the Jacobian matrix d Heating[lay=i] / d T[lay=j]
        """
        net=self.spectral_integration(self.flux_net_nu)
        self.kernel=np.empty((self.Nlay,self.Nlay))
        tlay=self.tlay
        dT = epsilon*tlay
        self.tlay = tlay + dT
        newpiBatm = self.source_function(integral=integral)

        for ilay in range(self.Nlay):
            pibatm = np.copy(self.piBatm)
            pibatm[ilay] = newpiBatm[ilay]
            _, _, flux_net_tmp = \
                solve_2stream_nu(pibatm, self.dtau, self.single_scat_albedo, self.asym_param,
                    self.flux_top_dw_nu, mu0 = mu0, flux_at_level=flux_at_level)
            net_tmp = self.spectral_integration(flux_net_tmp)
            self.kernel[ilay]=(net-net_tmp)/dT[ilay]
        self.kernel[:,:-1]-=self.kernel[:,1:]
        self.tlay=tlay
        self.tlay_kernel=self.tlay
        if per_unit_mass: self.kernel*=self.inv_dmass

    def flux_divergence(self, per_unit_mass = True,
            compute_kernel=False, **kwargs):
        """Computes the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
            net: array
                Net fluxes at level surfaces
        """
        if self.flux_net_nu is None:
            raise RuntimeError('should have ran emission_spectrum_2stream.')
        net = self.spectral_integration(self.flux_net_nu)
        H = -np.copy(net)
        H[:-1] -= H[1:]
        H[-1] += self.internal_flux
        if per_unit_mass: H *= self.inv_dmass
        if compute_kernel: self.H_kernel = H
        return H, net

    def heating_rate(self, compute_kernel=False, dTmax_use_kernel=None, **kwargs):
        if (not compute_kernel) and (dTmax_use_kernel is not None):
            dT=self.tlay-self.tlay_kernel
            if np.amax(np.abs(dT)) < dTmax_use_kernel:
                try:
                    H = self.H_kernel + np.dot(dT,self.kernel)
                except:
                    raise RuntimeError("Kernel has not been precomputed")
                net = np.zeros_like(H)
                return H, net
        _ = self.emission_spectrum_2stream(flux_at_level=True, integral=True,
                compute_kernel=compute_kernel, **kwargs)
        H, net = self.flux_divergence(compute_kernel=compute_kernel, **kwargs)
        if compute_kernel:
            self.H_kernel = H
            self.tau_rad = 1./np.amax(np.abs(self.kernel.diagonal()))
        return H, net


    def bolometric_fluxes(self, per_unit_mass = True):
        """Computes the bolometric fluxes at levels and the divergence of the net flux in the layers
        (used to compute heating rates).

        :func:`emission_spectrum_2stream` needs to be ran first.

        Parameters
        ----------
            per_unit_mass: bool
                If True, the heating rates are normalized by the
                mass of each layer (result in W/kg).

        Returns
        -------
            up: array
                Upward fluxes at level surfaces
            dw: array
                Downward fluxes at level surfaces
            net: array
                Net fluxes at level surfaces
            H: array
                Heating rate in each layer (Difference of the net fluxes). Positive means heating.
                The last value is the net flux impinging on the surface + the internal flux.
        """
        H, net = self.flux_divergence(per_unit_mass = per_unit_mass)
        up=self.spectral_integration(self.flux_up_nu)
        dw=self.spectral_integration(self.flux_down_nu)
        return up, dw, net, H

    def transmittance_profile(self, **kwargs):
        """Computes the transmittance profile of an atmosphere,
        i.e. Exp(-tau) for each layer of the model.
        Real work done in the numbafied function path_integral_corrk/xsec
        depending on the type of data.
        """
        self.opacity(**kwargs)
        self.compute_tangent_path()
        self.compute_density()
        if self.Ng is not None:
            self.weights=self.kdatabase.weights
            transmittance=path_integral_corrk( \
                self.Nlay-1,self.Nw,self.Ng,self.tangent_path,self.density,self.kdata,self.weights)
        else:
            transmittance=path_integral_xsec( \
                self.Nlay-1,self.Nw,self.tangent_path,self.density,self.kdata)
        return transmittance

    def transmission_spectrum(self, normalized=False, Rstar=None, **kwargs):
        r"""Computes the transmission spectrum of the atmosphere.
        In general (see options below), the code returns the transit depth:

        .. math::
            \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2),

        where

        .. math::
          \alpha_\nu=2 \pi \int_0^{z_{max}} (R_p+z)*(1-e^{-\tau_\nu(z)) d z.
        
        Parameters
        ----------
            Rstar: float or astropy.unit object, optional
                Radius of the host star. If a float is specified, meters are assumed.
                Does not need to be given here if
                it has already been specified as an attribute of the :class:`Atm` object.
                If specified, the result is the transit depth:

                .. math::
                  \delta_\nu=(\pi R_p^2+\alpha_\nu)/(\pi R_{star}^2).

            normalized: boolean, optional
                Used only if self.Rstar and Rstar are None:

                * If True,
                  the result is normalized to the planetary radius:

                  .. math::
                    \delta_\nu=1+\frac{\alpha_\nu}{\pi R_p^2}.
                * If False,
                                    
                  .. math::
                    \delta_\nu=\pi R_p^2+\alpha_\nu.

        Returns
        -------
            array
                The transit spectrum (see above for normalization options).
        """
        self.set_Rstar(Rstar)
        transmittance=self.transmittance_profile(**kwargs)
        self.compute_area()
        res=Spectrum((np.dot(self.area,(1.-transmittance))),self.wns,self.wnedges)
        if self.Rstar is not None:
            return (res+(PI*self.Rp**2))/(PI*self.Rstar**2)
        elif normalized:
            return res/(PI*self.Rp**2)+1
        else:
            return res+(PI*self.Rp**2)

    def __repr__(self):
        """Method to output header
        """
        output=super().__repr__()
        output+="""
    k_database      :
        {kdatab}
    cia_database    :
        {cdatab}""".format(kdatab=self.kdatabase, cdatab=self.gas_mix.cia_database)
        if self.gas_mix._wn_range is not None:
            output+='    wn range        : '+ self.gas_mix._wn_range +'\n'

        return output
        
    def exp_minus_tau(self):
        """Sums Exp(-tau) over gauss points
        """
        weights=self.kdatabase.weights
        return np.sum(np.exp(-self.tau[1:])*weights,axis=2)

    def exp_minus_tau_g(self, g_index):
        """Sums Exp(-tau) over gauss point
        """
        return np.exp(-self.tau[1:,:,g_index])

    def blackbody(self, layer_idx=-1, integral=True):
        """Computes the surface black body flux (in W/m^2/cm^-1) for the temperature
        of layer `layer_idx`.

        Parameters
        ----------
            layer_idx; int
                Index of layer used for the temperature.
            integral: boolean, optional
                * If true, the black body is integrated within each wavenumber bin.
                * If not, only the central value is used.
                  False is faster and should be ok for small bins,
                  but True is the correct version. 
        Returns
        -------
            Spectrum object
                Spectral flux in W/m^2/cm^-1
        """
        if integral:
            piBatm=PI*Bnu_integral_num(self.wnedges,self.tlay[layer_idx])/np.diff(self.wnedges)
        else:
            piBatm=PI*Bnu(self.wns[:],self.tlay[layer_idx])
        return Spectrum(piBatm,self.wns,self.wnedges)

    def surf_bb(self, **kwargs):
        """Computes the surface black body flux (in W/m^2/cm^-1).

        See :any:`blackbody` for options.

        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the surface (in W/m^2/cm^-1)
        """
        return self.blackbody(layer_idx=-1, **kwargs)

    def top_bb(self, **kwargs):
        """Computes the top of atmosphere black body flux (in W/m^2/cm^-1).

        See :any:`blackbody` for options.

        Returns
        -------
            Spectrum object
                Spectral flux of a bb at the temperature at the top of atmosphere (in W/m^2/cm^-1)
        """
        return self.blackbody(layer_idx=0, **kwargs)




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A module to handle ouputs rebinning and plotting
"""
import numpy as np
import h5py
from exo_k.util.interp import rebin
from exo_k.util.spectral_object import Spectral_object

class Spectrum(Spectral_object):
    """A class defining a Spectrum object to plot and manipulate.
    """

    def __init__(self, value=None, wns=None, wnedges=None, filename=None,
            from_taurex=False, dataset='native_spectrum', **kwargs):
        """Instanciate with a value, bin centers, and bin edges.
        Can also load a Taurex spectrum if filename is provided.
        """
        self.wn_unit='cm^-1'
        self.value=value
        self.wns=wns
        self.wnedges=wnedges
        if filename is not None:
            if from_taurex:
                self.load_taurex(filename, dataset)
            elif filename.lower().endswith(('.hdf5', '.h5')):
                self.read_hdf5(filename)
            elif filename.lower().endswith(('.dat', '.txt')):
                self.read_ascii(filename, **kwargs)
        if (self.wnedges is None) and (self.wns is not None):
            self.wnedges=np.concatenate(([self.wns[0]],
                (self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))

    
    def copy(self):
        """Deep copy of the spectrum.
        """
        return Spectrum(self.value,self.wns,self.wnedges)

    def plot_spectrum(self, ax, per_wavenumber=True, x_axis='wls',
            xscale=None, yscale=None, **kwarg):
        """Plot the spectrum
        
        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            per_wavenumber: bool, optional
                Defines the units of spectral flux density.
                False converts to per wavelength units.
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        if per_wavenumber:
            toplot=self.value
        else:
            toplot=self.value/self.wls**2*1.e4
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.set_ylabel('Flux')
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def bin_down(self, wnedges):
        """Bins down the spectrum to a new grid of wnedges by conserving area.
        
        Parameters
        ----------
            wnedges: array
                Wavenumbers of the bin edges to be used
        """
        wnedges=np.array(wnedges)
        self.value=rebin(self.value,self.wnedges,wnedges)
        self.wnedges=wnedges
        self.wns=0.5*(self.wnedges[:-1]+self.wnedges[1:])

    def bin_down_cp(self, wnedges):
        """Returns a new binned down spectrum to a grid of wnedges by conserving area.
        
        Parameters
        ----------
            wnedges: array
                Wavenumbers of the bin edges to be used

        Returns
        -------
            :class:`Spectrum`
                Binned down spectrum
        """
        res=self.copy()
        res.bin_down(wnedges)
        return res

    def __add__(self,other):
        """Defines addition
        """
        if (isinstance(other,float) or isinstance(other,int)):
            return Spectrum(self.value+other,self.wns,self.wnedges)
        elif (self.wns.size==other.wns.size) and np.array_equal(self.wns,other.wns):
            val=self.value+other.value
            return Spectrum(val,self.wns,self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __sub__(self,other):
        """Defines substraction
        """
        if (isinstance(other,float) or isinstance(other,int)):
            return Spectrum(self.value-other,self.wns,self.wnedges)
        elif (self.wns.size==other.wns.size) and np.array_equal(self.wns,other.wns):
            val=self.value-other.value
            return Spectrum(val,self.wns,self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __mul__(self,other):
        """Defines multiplication
        """
        if (isinstance(other,float) or isinstance(other,int)):
            return Spectrum(self.value*other,self.wns,self.wnedges)
        elif (self.wns.size==other.wns.size) and np.array_equal(self.wns,other.wns):
            val=self.value*other.value
            return Spectrum(val,self.wns,self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def __truediv__(self,other):
        """Defines division
        """
        if (isinstance(other,float) or isinstance(other,int)):
            return Spectrum(self.value/other,self.wns,self.wnedges)
        elif (self.wns.size==other.wns.size) and np.array_equal(self.wns,other.wns):
            val=self.value/other.value
            return Spectrum(val,self.wns,self.wnedges)
        else:
            raise RuntimeError('The two spectra do not have the same spectral sampling.')

    def std(self):
        """Defines standard deviation
        """
        return self.value.std()

    def abs(self):
        """Defines absolute value
        """
        return Spectrum(np.abs(self.value),self.wns,self.wnedges)

    def log10(self):
        """Defines Log 10
        """
        return Spectrum(np.log10(self.value),self.wns,self.wnedges)

    @property
    def total(self):
        """Defines the weighted sum over the spectrum
        """
        dw=np.diff(self.wnedges)
        return np.dot(self.value,dw)

    def read_hdf5(self, filename=None):
        """Reads data in a hdf5 file

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        if (filename is None or not filename.lower().endswith(('.hdf5', '.h5'))):
            raise RuntimeError("You should provide an input hdf5 file")
        f = h5py.File(filename, 'r')
        self.wns=f['bin_centers'][...]
        self.wnedges=f['bin_edges'][...]
        if 'units' in f['bin_edges'].attrs:
            self.wn_unit=f['bin_edges'].attrs['units']
        else:
            if 'units' in f['bin_centers'].attrs:
                self.wn_unit=f['bin_centers'].attrs['units']
        self.value=f['spectrum'][...]
        f.close()  

    def write_hdf5(self, filename):
        """Saves data in a hdf5 format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.hdf5', '.h5')):
            fullfilename=filename+'.h5'
        compression="gzip"
        f = h5py.File(fullfilename, 'w')
        f.create_dataset("spectrum", data=self.value, compression=compression)
        f.create_dataset("bin_edges", data=self.wnedges, compression=compression)
        f["bin_edges"].attrs["units"] = 'cm^-1'
        f.create_dataset("bin_centers", data=self.wns, compression=compression)
        f["bin_centers"].attrs["units"] = 'cm^-1'
        f.close()

    def read_ascii(self, filename, spec_axis='wns', skip_header=None):
        """Saves data in a ascii format

        Parameters
        ----------
            filename: str
                Name of the file to be read
            spec_axis: str
                Whether the secptral axis in the file is
                wavenumber in cm^-1 ('wns') or wavelength in microns ('wls')
            skip_header: int
                Number of lines to skip
        """
        raw=np.genfromtxt(filename, skip_header=skip_header,
            usecols=(0,1), names=('spec_axis','value'))
        if spec_axis=='wls':
            raw['spec_axis']=1.e4/raw['spec_axis'][::-1] #conversion to wavenumber in cm-1
            raw['value']=raw['value'][::-1]
        self.value=raw['value']
        self.wns=raw['spec_axis']

    def write_ascii(self, filename, fmt='%.18e', spec_axis='wns', header=None):
        """Saves data in a ascii format

        Parameters
        ----------
            filename: str
                Name of the file to be created and saved
        """
        fullfilename=filename
        if not filename.lower().endswith(('.dat', '.txt')):
            fullfilename=filename+'.dat'
        head=header
        if spec_axis=='wns':
            if head is None: head='wavenumber(cm^-1)     spectrum'
            np.savetxt(fullfilename, np.array([self.wns,self.value]).transpose(),
                fmt=fmt, header=head)
        else:    
            if head is None: head='wavelength(micron)    spectrum'
            np.savetxt(fullfilename, np.array([self.wls[::-1],self.value[::-1]]).transpose(),
                fmt=fmt, header=head)

    def load_taurex(self, filename,dataset='native_spectrum'):
        """Loads a taurex file

        Parameters
        ----------
            filename: str
                Full name (path) of the hdf5 file to load
            dataset: str
                Name of the hdf5 dataset to load
        """
        f = h5py.File(filename, 'r')
        self.wns=f['Output/Spectra/native_wngrid'][...]
        self.value=f['Output/Spectra/'+dataset][...]
        f.close()
        self.wnedges=np.concatenate(([self.wns[0]],(self.wns[:-1]+self.wns[1:])*0.5,[self.wns[-1]]))


    def __repr__(self):
        """Method to output header
        """
        output="""
        value        : {val}
        wl (microns) : {wl}
        """.format(val=self.value,wl=self.wls)
        return output




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte
"""
import numpy as np
from .util.molar_mass import Molar_mass
from .util.interp import RandOverlap_2_kdata_prof, rm_molec
from .rayleigh import Rayleigh
from .util.spectral_object import Spectral_object
from .util.singleton import Singleton

class Gas_mix(Spectral_object):
    """Dict-like class to handle gas composition (with background gas) and molar mass.

    If `logp_array`, `t_array`, and radiative databases are provided, :any:`cross_section`
    can be used to compute the opacity of the gas
    """

    def __init__(self, composition={}, logp_array=None, t_array=None,
        k_database=None, cia_database=None):
        """__init_ Instantiates
        a Gas_mix object and computes the vmr of the 'background' gas.
        """
        self.Narray=None
        self._wn_range=None
        self.iw_min=None
        self.iw_max=None
        self.wns=None
        self.wnedges=None
        self.Nw=None
        self.need_to_recompute_vmr_array = True

        self.set_composition(composition)
        self.set_logPT(logp_array=logp_array, t_array=t_array)
        self.set_k_database(k_database)
        self.set_cia_database(cia_database)


    def set_composition(self, composition):
        """Reset composition and computes the vmr of the 'background' gas.

        Parameters
        ----------
            composition: dict
                Keys are molecule names. Values are vmr or arrays of vmr.
            bg_gas: str
                Name of the background molecule. 
                If None, it is inferred from the molecule for which vmr='background'.
        """
        self.composition = dict()
        self.bg_gas = None
        for mol,vmr in composition.items():
            if isinstance(vmr, list):
                self.composition[mol]=np.array(vmr)
            elif isinstance(vmr, str):
                self.bg_gas = mol
            else:
                self.composition[mol] = vmr
        if (self.bg_gas is not None):
            self.composition['inactive_gas'] = 0.
        else:
            self.composition['inactive_gas'] = 'background'
            self.bg_gas = 'inactive_gas'
        self.get_background_vmr()
        self.need_to_recompute_vmr_array = True

    def set_logPT(self, logp_array=None, t_array=None):
        """Sets the pressure (in Pa) and temperature fields.
        """
        if logp_array is not None:
            if hasattr(logp_array, "__len__"):
                self.logp_array=np.array(logp_array, dtype=float)
            else:
                self.logp_array=np.array([logp_array], dtype=float)

            if hasattr(t_array, "__len__"):
                self.t_array=np.array(t_array)
            else:
                self.t_array=np.array([t_array])

            if self.logp_array.size != self.t_array.size:
                raise TypeError(\
                    'logp_array and t_array should be 1d lists or arrays of the same size')
            self.Narray=self.logp_array.size

    def get_background_vmr(self):
        """Computes the volume mixing ratio of the background gas in a mix
        Uses the fact that self.composition is a dictionary
        with the name of the molecules as keys and the vmr as values.
        vmr are either float or arrays.
        
        At this stage, the background gas should be identified by `self.bg_gas`,
        and its Vol. Mix. Ratio will be updated in the composition dict. 
        """
        other_vmr=0.
        for mol,vmr in self.composition.items():
            if mol==self.bg_gas:
                continue
            try:
                other_vmr+=vmr
            except ValueError:
                raise TypeError('Incompatible shapes in Gas_mix arrays.')
        if np.amax(other_vmr)>1.:
            print("""Careful: the sum of the vmr of your gas components is > 1.
            If there is a background gas, its vmr will become negative.
            I hope you know what you are doing.""")
        self.composition[self.bg_gas]=1.-other_vmr

    def molar_mass(self):
        """Computes and returns the molar mass of a mix of gases

        Returns
        -------
            float or array:
                Molar mass of the active gases in kg/mol
        """
        mol_mass_active_gases=0.
        vmr_active_gases=0.

        for mol,vmr in self.composition.items():
            if mol!='inactive_gas':
                Mmol=Molar_mass().fetch(mol)
                mol_mass_active_gases+=vmr*Mmol
                vmr_active_gases+=vmr
        mol_mass=mol_mass_active_gases/vmr_active_gases     
        return mol_mass

    def get_vmr_array(self, sh=None):
        """Returns a dictionary with an array of vol. mix. ratios for each species. 

        Parameters
        ----------
            sh: set or list
                shape of the array wanted if all the vmr are floats.
                If some are already arrays, check whether the shape is the correct one. 

        Returns
        -------
            vmr_array: dict
                A dictionary with the an array of vmr per species.
            cst_array: boolean
                Is True if all the values in the arrays are constant.
        """
        vmr_array=dict()
        cst_array=True
        for mol,vmr in self.composition.items():
            if isinstance(vmr,(float,int)):
                vmr_array[mol]=np.ones(sh)*vmr
            else:
                cst_array=False
                vmr_array[mol]=np.array(vmr) # this np.array could probably go because vmr should be an array at this stage. 
                if not np.array_equal(vmr_array[mol].shape, sh):
                    print(vmr_array[mol].shape)
                    raise RuntimeError('Wrong shape in get_vmr_array')
        return vmr_array, cst_array

    def get_vmr_array_basic_molecules(self, sh=None):
        """Returns a dictionary with an array of vol. mix. ratios for each and decomposes
        gas mixtures into basic molecules. 

        Parameters
        ----------

        Returns
        -------
            vmr_array_basic_mol: dictionary of arrays
                A dictionary with the an array of vmr per species.
                Each species is considered a basic molecule.
            vmr_array: dictionary of arrays
                A dictionary with the an array of vmr per species.
                Species in this dictionary can be a gas mixture. 
            cst_array: boolean
                Is True if all the values in the arrays are constant.
        """
        vmr_array_basic_mol=dict()
        cst_array=True
        vmr_array, tmp_cst_array = self.get_vmr_array(sh=sh)
        if not tmp_cst_array: cst_array=False
        for mol,vmr in vmr_array.items():
            if mol in Known_composite_species().keys():
                tmp, _, tmp_cst_array= \
                    Known_composite_species()[mol].get_vmr_array_basic_molecules(sh=sh)
                if not tmp_cst_array: cst_array=False
                for mol2,vmr2 in tmp.items():
                    if mol2 in vmr_array_basic_mol.keys():
                        vmr_array_basic_mol[mol2]+=vmr*vmr2
                    else:
                        vmr_array_basic_mol[mol2]=vmr*vmr2
            else:
                if mol in vmr_array_basic_mol.keys():
                    vmr_array_basic_mol[mol]+=vmr
                else:
                    vmr_array_basic_mol[mol]=vmr
        return vmr_array_basic_mol, vmr_array, cst_array


    def set_k_database(self, k_database=None):
        """Change the radiative database attached to the current instance of Gas_mix

        Parameters
        ----------
            k_database: :class:`~exo_k.kdatabase.Kdatabase` object
                New Kdatabase to use.
        """
        self.kdatabase=k_database
        if self.kdatabase is None:
            self.Ng=None
        else:
            self.Ng=self.kdatabase.Ng
            if self.kdatabase.p_unit != 'Pa' or rm_molec(self.kdatabase.kdata_unit) != 'm^2':
                print("""You're being Bad!!! You are trying *NOT* to use MKS units!!!
                You can convert to mks using convert_to_mks on your Kdatabase.
                More generally, you can specify exo_k.Settings().set_mks(True) 
                to set MKS system as default for all newly loaded data,
                but beware that this global setting is overridden by local options
                specified during the loading.
                You will have to reload all your data though.
                (A good thing it does not take so long). """)
                raise RuntimeError("Bad units in the Kdatabase used with Gas_mix.")
            if (not self.kdatabase.consolidated_p_unit) \
                or (not self.kdatabase.consolidated_kdata_unit):
                raise RuntimeError( \
                    """All tables in the database should have the same units to proceed.
                    You should probably use convert_to_mks().""")

    def set_cia_database(self, cia_database=None):
        """Changes the CIA database attached to the current instance of Gas_mix

        Parameters
        ----------
            cia_database: :class:`~exo_k.ciadatabase.CIAdatabase` object
                New CIAdatabase to use.
        """
        self.cia_database=cia_database
        if self.cia_database is not None:
            if self.cia_database.abs_coeff_unit != 'm^5':
                print("""You're being Bad!!! You are trying *NOT* to use MKS units!!!
                You can convert to mks using convert_to_mks on your CIAdatabase.
                More generally, you can specify exo_k.Settings().set_mks(True) 
                to set MKS system as default for all newly loaded data,
                but beware that this global setting is overridden by local options
                specified during the loading.
                You will have to reload all your data though.
                (A good thing it does not take so long). """)

                raise RuntimeError("Bad units in the CIAdatabase used with Gas_mix.")

    def _compute_spectral_range(self, wn_range=None, wl_range=None):
        """Converts an unordered spectral range in either wavenumber or wavelength
        in an ordered wavenumber range.

        Parameters
        ----------
            wn_range: list or array of size 2
                Minimum and maximum wavenumber (in cm^-1).
            wl_range: list or array of size 2
                Minimum and maximum wavelength (in micron)
        """
        if wl_range is not None:
            if wn_range is not None:
                print('Cannot specify both wl and wn range!')
                raise RuntimeError()
            else:
                _wn_range=np.sort(10000./np.array(wl_range))
        else:
            if wn_range is not None:
                _wn_range=np.sort(np.array(wn_range))
            else:
                _wn_range=self._wn_range
        return _wn_range

    def set_spectral_range(self, wn_range=None, wl_range=None):
        """Sets the default spectral range in which computations will be done by specifying
        either the wavenumber or the wavelength range.

        Parameters
        ----------
            wn_range: list or array of size 2
                Minimum and maximum wavenumber (in cm^-1).
            wl_range: list or array of size 2
                Minimum and maximum wavelength (in micron)
        """
        self._wn_range=self._compute_spectral_range(wn_range=wn_range, wl_range=wl_range)

    def _compute_wn_range_indices(self, wn_range=None):
        """Compute the starting and ending indices to be used for current wn_range
        """
        if wn_range is None:
            local_wn_range=self._wn_range
        else:
            local_wn_range=wn_range
        if local_wn_range is None:
            self.iw_min=0
            self.iw_max=self.kdatabase.Nw
        else:
            self.iw_min, self.iw_max = np.where((self.kdatabase.wnedges > local_wn_range[0]) \
                & (self.kdatabase.wnedges <= local_wn_range[1]))[0][[0,-1]]
            # to be consistent with interpolate_kdata

    def cross_section(self, composition=None, logp_array=None, t_array=None,
            wl_range=None, wn_range=None, rayleigh=True,
            write=0, random_overlap=False, logp_interp=True, use_basic_molecules=False, **kwargs):
        """Computes the cross section (m^2/total number of molecule) for the mix
        at each of the logPT points as a function of wavenumber (and possibly g point).

        Parameters
        ----------
            wl_range: array or list of two values, optional
                Wavelength range to cover.
            wn_range: array or list of two values, optional
                Wavenumber range to cover.
            rayleigh: boolean, optional
                Whether to compute rayleigh scattering.
            random_overlap: boolean, optional
                Whether Ktable opacities are added linearly (False),
                or using random overlap method (True).

        Returns
        -------
            kdata_array: array
                Cross section array of shape (layer number, Nw (, Ng if corrk)).

        After every computation, the following variables are updated to account for any possible
        change in spectral range:

          * self.Nw, Number of wavenumber bins
          * self.wns, Wavenumber array
          * self.wnedges, Wavenumber of the edges of the bins
        """
        if self.kdatabase is None: raise RuntimeError("""k_database not provided. 
        Use the k_database keyword during initialization or use the set_k_database method.""")
        if not self.kdatabase.consolidated_wn_grid: raise RuntimeError("""
            All tables in the database should have the same wavenumber grid to proceed.
            You should probably use bin_down().""")
        if self.cia_database is not None and \
            (not np.array_equal(self.cia_database.wns,self.kdatabase.wns)):
            raise RuntimeError("""CIAdatabase not sampled on the right wavenumber grid.
            You should probably run something like CIAdatabase.sample(Kdatabase.wns).""")
        
        self.set_logPT(logp_array=logp_array, t_array=t_array)
        if self.Narray is None:
            raise RuntimeError('You must prescribe logP (in Pa) and T arrays first.')

        if composition is not None:
            self.set_composition(composition)
        if self.need_to_recompute_vmr_array:
            self.vmr_arr_base_mol, self.vmr_arr, self.cst_array = \
                self.get_vmr_array_basic_molecules((self.Narray,))
            self.need_to_recompute_vmr_array = False
            if write>6 :
                print('recomputed vmr array:', self.vmr_arr_base_mol, self.vmr_arr) 
        if use_basic_molecules:
            molecs=self.vmr_arr_base_mol.keys()
        else:
            molecs=self.vmr_arr.keys()
        if write>6 :
            print('molecs:', molecs) 
        mol_to_be_done=list(set(molecs).intersection(self.kdatabase.molecules))
        if not mol_to_be_done:
            if 'total' in self.kdatabase.molecules:
                #special case where you have one ktable that describes the whole gas.
                mol_to_be_done=['total']
                use_basic_molecules=False
                self.vmr_arr['total']=np.ones((self.Narray))
            else:
                raise RuntimeError("""The k_database you provided 
                    should contain at least one molecule in your atm,
                    or a ktable with the 'total' key.""")
        if all(elem in self.kdatabase.molecules for elem in molecs):
            if write>3 : print("""I have all the molecules present in the atmosphere
              in ktables provided:""")
        else:
            if write>3 : print("""Some missing molecules in my database,
             I ll compute opacites with the available ones:""")
        if write>3 : print(mol_to_be_done)

        # does what needs to be done to reduce the spectral range
        local_wn_range=self._compute_spectral_range(wl_range=wl_range, wn_range=wn_range)
        self._compute_wn_range_indices(wn_range=local_wn_range)
        self.wnedges=np.copy(self.kdatabase.wnedges[self.iw_min:self.iw_max+1])
        self.dwnedges=np.diff(self.wnedges)
        self.wns=np.copy(self.kdatabase.wns[self.iw_min:self.iw_max]) # 2021: are these copy needed?
        self.Nw=self.wns.size

        first_mol=True
        if use_basic_molecules:
            vmr_arr_to_use = self.vmr_arr_base_mol
        else:
            vmr_arr_to_use = self.vmr_arr
        for mol in mol_to_be_done:
            tmp_kdata=self.kdatabase[mol].interpolate_kdata(logp_array=self.logp_array,
                t_array=self.t_array, x_array=vmr_arr_to_use[mol], wngrid_limit=local_wn_range,
                logp_interp=logp_interp,
                **kwargs)
            if first_mol:
                kdata_array=tmp_kdata
                first_mol=False
            else:
                if random_overlap and (self.kdatabase.Ng is not None):
                    kdata_array=RandOverlap_2_kdata_prof(self.Narray,
                        self.Nw, self.kdatabase.Ng, 
                        kdata_array,tmp_kdata, self.kdatabase.weights,
                        self.kdatabase.ggrid)
                else:
                    kdata_array+=tmp_kdata

        if rayleigh or (self.cia_database is not None):
            # continua always use base molecules
            cont_sig=np.zeros((self.Narray,self.Nw))
            if rayleigh:
                if self.cst_array:
                    cont_sig+=Rayleigh().sigma(self.wns, self.vmr_arr_base_mol)
                else:
                    cont_sig+=Rayleigh().sigma_array(self.wns, self.vmr_arr_base_mol)
                self.kdata_scat=np.copy(cont_sig)
            if self.cia_database is not None:
                cont_sig+=self.cia_database.cia_cross_section(self.logp_array,
                    self.t_array, self.vmr_arr_base_mol, wngrid_limit=local_wn_range, Nw=self.Nw)
            if self.kdatabase.Ng is None:
                kdata_array+=cont_sig
            else:
                kdata_array+=cont_sig[:,:,None]

        return kdata_array
    
    def __getitem__(self, molecule):
        """Overrides getitem
        """
        return self.composition[molecule]

    def __setitem__(self, molecule, vmr):
        """Overrides setitem and recomputes the background gas mixing ratio if needed
        """
        if isinstance(vmr, str):
            self.bg_gas=molecule
            self.composition['inactive_gas']=0.
        else:
            if isinstance(vmr, list):
                self.composition[molecule]=np.array(vmr)
            else:
                self.composition[molecule]=vmr
            if molecule==self.bg_gas:
                self.bg_gas='inactive_gas'
        self.get_background_vmr()
    
    def items(self):
        """Emulates dict.items() method
        """
        return self.composition.items()

    def values(self):
        """Emulates dict.values() method
        """
        return self.composition.values()

    def keys(self):
        """Emulates dict.keys() method
        """
        return self.composition.keys()

    def __repr__(self):
        """Method to output
        """
        output='Volume mixing ratios in Gas_mix: \n'
        for mol, vmr in self.composition.items():
            output+=mol+'->'+str(vmr)+'\n'
        return output

    def copy(self):
        """Deep copy of the dict and arrays. 
        The databases are not deep copied. 
        """
        res=Gas_mix(self.composition, bg_gas=self.bg_gas,
            k_database=self.kdatabase, cia_database=self.cia_database)
        res.logp_array=np.copy(self.logp_array)
        res.t_array=np.copy(self.t_array)
        res.Narray=self.Narray
        res._wn_range=np.copy(self._wn_range)
        res.iw_min=self.iw_min
        res.iw_max=self.iw_max

    def mix_with(self, other_gas, vmr_other_gas):
        """Mix with other Gas_mix.
        """
        raise NotImplementedError()

    def __mul__(self, vmr):    
        """Defines multiplication
        """
        raise NotImplementedError()
        #composition=dict()
        #for mol,mol_vmr in self.composition.items():
        #    composition[mol]=vmr*mol_vmr
        #res=Gas_mix(composition, bg_gas=self.bg_gas)
        #return res

    __rmul__ = __mul__


class Known_composite_species(Singleton):
    """A class to store composite gas mixes that can be used in other gas mixes
    This class can also store the molar mass of custom gases with arbitrary names
    (for example: My_gas, earth_background).
    """
    
    def init(self, *args, **kwds):
        """Initializes empty dictionary of custom molecular masses.
        """
        self._known_composite_species={}

    def add_species(self, species_dict):
        """Add one or several composite species to the database. 

        Parameters
        ----------
            species_dict: dict
                Keys are gases names (they do not have to be real molecules).
                Values are a gas_mix object. 
        """
        self._known_composite_species.update(species_dict)
        for species, gas_mix in species_dict.items():
            Molar_mass().add_species({species: gas_mix.molar_mass()})

    def items(self):
        """Emulates dict.items() method
        """
        return self._known_composite_species.items()

    def values(self):
        """Emulates dict.values() method
        """
        return self._known_composite_species.values()

    def keys(self):
        """Emulates dict.keys() method
        """
        return self._known_composite_species.keys()

    def __getitem__(self, molecule):
        """Overrides getitem
        """
        return self._known_composite_species[molecule]

    def __setitem__(self, molecule, mix):
        """Overrides setitem
        """
        self._known_composite_species[molecule]=mix



# -*- coding: utf-8 -*-
"""
Created in Jan 2021

@author: jeremy leconte
"""
import numpy as np
from scipy.linalg import solve_banded
import numba

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream_nu_xsec(source_nu, dtau_nu, omega0_nu, g_asym_nu,
                flux_top_dw_nu, mu0=0.5, alb_surf=0., flux_at_level=False):
    """Deals with the spectral axis
    """
    NLEV, NW = source_nu.shape
    flux_up=np.empty((NLEV, NW))
    flux_dw=np.empty((NLEV, NW))
    flux_net=np.empty((NLEV, NW))
    for iW in range(NW):
        flux_up[:,iW], flux_dw[:,iW], flux_net[:,iW] = \
            solve_2stream(source_nu[:,iW], dtau_nu[:,iW],
                omega0_nu[:,iW], g_asym_nu[:,iW],
                mu0=mu0, flux_top_dw=flux_top_dw_nu[iW],
                alb_surf=alb_surf, flux_at_level=flux_at_level)
    return flux_up, flux_dw, flux_net

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream_nu_corrk(source_nu, dtau_nu, omega0_nu, g_asym_nu,
                flux_top_dw_nu, mu0=0.5, alb_surf=0., flux_at_level=False):
    """Deals with the spectral axis
    """
    NLEV, NW = source_nu.shape
    NG = dtau_nu.shape[-1]
    flux_up=np.empty((NLEV, NW, NG))
    flux_dw=np.empty((NLEV, NW, NG))
    flux_net=np.empty((NLEV, NW, NG))
    for iW in range(NW):
        for iG in range(NG):
            flux_up[:,iW, iG], flux_dw[:,iW, iG], flux_net[:,iW, iG] = \
                solve_2stream(source_nu[:,iW], dtau_nu[:,iW,iG], 
                    omega0_nu[:,iW,iG], g_asym_nu[:,iW,iG],
                    mu0=mu0, flux_top_dw=flux_top_dw_nu[iW],
                    alb_surf=alb_surf, flux_at_level=flux_at_level)
    return flux_up, flux_dw, flux_net

@numba.jit(nopython=True, fastmath=True, cache=True)
def solve_2stream(source, dtau, omega0, g_asym,
                mu0=0.5, flux_top_dw=0., alb_surf=0., flux_at_level=False):
    """After Toon et al. (JGR, 1989). Equation number refer to this article.
    
    emis_surf=1.-alb_surf
    
    As we only consider hemispheric mean or quadrature, mu1==mu0

    Parameters
    ----------
        source: array
            pi*B (Planck function) at each of the Nlay+1 level interfaces of the model
            (top to bottom). Thanks to the pi factor it can be compared to the incoming
            flux.
        dtau: array
            optical depth of each of the Nlay model layers.
        omega0: float or array
            single scattering albedo of each layer
        g_asym: float or array
            asymmetry factor (g in Toon et al.)
        mu0: float
            1./2. yields hemisperic mean approx
            1./sqrt(3) yields quadrature
        flux_top_dw: float
            Incoming difuse flux at the upper boundary
        alb_surf: float
            Surface albedo. Emissivity is assumed to be 1.-alb_surf
        flux_at_level: bool, optional
            if flux_at_level is True, fluxes are calculated at the level surfaces.
            If False, fluxes are computed at the middle of the layers.
            The top of atmosphere flux is always computed at the top of the uppermost layer
            (1st level).


    """
    Nlay=dtau.size
    gam_1, gam_2=_gammas_toon(omega0, g_asym, mu0=mu0)
    flux_net, J4pimu=matrix_toon_tridiag(Nlay, source, dtau, gam_1, gam_2, mu0,
                flux_top_dw, alb_surf, flux_at_level)
    return 0.5*(J4pimu+flux_net), 0.5*(J4pimu-flux_net), flux_net

@numba.jit(nopython=True, fastmath=True, cache=True)
def _gammas_toon(omega0, g_asym, mu0=0.5):
    """mu0=0.5 yields hemispheric mean
    
    Parameters
    ----------
        omega0: float or array
            single scattering albedo of each layer
        g_asym: float or array
            asymmetry factor (g in Toon et al.)
        mu0: float
            1./2. yields hemisperic mean approx
            1./sqrt(3) yields quadrature

    Returns
    -------
        gamma1, gamma2: floats or arrays
            gammas defined in Table 1 of Toon et al. 
    """
    return (2.-omega0*(1.+g_asym))/(2.*mu0), omega0*(1.-g_asym)/(2.*mu0)


@numba.jit(nopython=True, fastmath=True, cache=True)
def matrix_toon_tridiag(Nlay, source, dtau, gam_1, gam_2, mu1,
        flux_top_dw, alb_surf, flux_at_level):
    """
    Returns
    -------
        flux_net: array
            Net flux at the bottom of the Nlay layers.
    """
    e_1, e_2, e_3, e_4 = e_i_toon(dtau, gam_1, gam_2)
    c_up_top, c_dw_top, c_up_bot, c_dw_bot = c_planck(source, dtau, gam_1, gam_2, mu1)
    A=np.empty((2*Nlay))
    B=np.empty((2*Nlay))
    D=np.empty((2*Nlay))
    E=np.empty((2*Nlay))
    # upper boundary
    A[0]=0.
    D[0]=-e_2[0]
    E[0]=flux_top_dw-c_dw_top[0]
    B[0]=e_1[0]
    #even in Toon numerotation. (remember python arrays start at 0.)
    A[1:-1:2]=e_1[:-1]*e_2[1:]-e_3[:-1]*e_4[1:]
    B[1:-1:2]=e_2[:-1]*e_2[1:]-e_4[:-1]*e_4[1:]
    D[1:-1:2]=e_1[1:]*e_4[1:]-e_2[1:]*e_3[1:]
    E[1:-1:2]=(c_up_top[1:]-c_up_bot[:-1])*e_2[1:]-(c_dw_top[1:]-c_dw_bot[:-1])*e_4[1:]
    # middle sign above different in my calculations and toon (+ in Toon)
    #odd
    A[2:-1:2]=e_2[:-1]*e_3[:-1]-e_4[:-1]*e_1[:-1]
    B[2:-1:2]=e_1[:-1]*e_1[1:]-e_3[:-1]*e_3[1:]
    D[2:-1:2]=e_3[:-1]*e_4[1:]-e_1[:-1]*e_2[1:]
    E[2:-1:2]=(c_up_top[1:]-c_up_bot[:-1])*e_3[:-1]-(c_dw_top[1:]-c_dw_bot[:-1])*e_1[:-1]
    #surface
    A[-1]=e_1[-1]-alb_surf*e_3[-1]
    B[-1]=e_2[-1]-alb_surf*e_4[-1]
    D[-1]=0.
    E[-1]=(1.-alb_surf)*source[-1]-c_up_bot[-1]+alb_surf*c_dw_bot[-1]
    #return mat,E
    Y=DTRIDGL(2*Nlay,A,B,D,E)
    flux_net = np.empty((Nlay+1))
    J4pimu   = np.empty((Nlay+1))
    if flux_at_level:
        c_up_mid, c_dw_mid = c_planck_mid(source, dtau, gam_1, gam_2, mu1)
        #print('mid:',np.amax(np.abs((c_up_mid-c_up_top)/c_up_top)))
        #print('bot:',np.amax(np.abs((c_up_bot-c_up_top)/c_up_top)))
        mid_factor1, mid_factor2 = mid_factor_toon(dtau, gam_1, gam_2)
        flux_net[1:] = Y[1::2]*mid_factor2+c_up_mid-c_dw_mid
        J4pimu[1:]   = Y[::2]*mid_factor1+c_up_mid+c_dw_mid
        #flux_net[-1] = Y[-2]*(e_1[-1]-e_3[-1])+Y[-1]*(e_2[-1]-e_4[-1])+c_up_bot[-1]-c_dw_bot[-1]
        #J4pimu[-1]   = Y[-2]*(e_1[-1]+e_3[-1])+Y[-1]*(e_2[-1]+e_4[-1])+c_up_bot[-1]+c_dw_bot[-1]
    else:
        flux_net[1:] = Y[::2]*(e_1-e_3)+Y[1::2]*(e_2-e_4)+c_up_bot-c_dw_bot
        J4pimu[1:]   = Y[::2]*(e_1+e_3)+Y[1::2]*(e_2+e_4)+c_up_bot+c_dw_bot
    flux_net[0]  = Y[0]*e_3[0]-Y[1]*e_4[0]+c_up_top[0]
    J4pimu[0]    = flux_net[0] + flux_top_dw
    flux_net[0] -= flux_top_dw
    return flux_net, J4pimu

@numba.jit(nopython=True, fastmath=True, cache=True)
def lambda_toon(gam_1, gam_2):
    """lambda from eq 21 of Toon et al.
    """
    return np.sqrt(gam_1*gam_1-gam_2*gam_2)

@numba.jit(nopython=True, fastmath=True, cache=True)
def lambda_GAMMA(gam_1, gam_2):
    """lambda and GAMMA from eq 21 and 22 of Toon et al.
    For GAMMA, the positive alterative is used (center in eq 22)
    """
    lamb=lambda_toon(gam_1, gam_2)
    GAMMA=gam_2/(lamb+gam_1)
    #GAMMA=(gam_1-lamb)/(gam_2)
    return lamb,GAMMA

@numba.jit(nopython=True, fastmath=True, cache=True)
def c_planck(source, dtau, gam_1, gam_2, mu1):
    """c_up/dw is for c+/- whithout direct beam scattering. 
    _top is for tau equal 0 (top of the layer)
    _bot is for tau=dtau (bottom of the layer)
    removed a pi factor because source is pi*B
    """
    #cst=2*mu1
    #cst=1. # this factor seems to avoid emissivity greater than one. 
           # but what is the analytical basis for this ???
    #print(gam_1+gam_2)
    inv_dtaugam=1./(dtau*(gam_1+gam_2))
    #print(1./(dtau*(gam_1+gam_2)), 1./dtau*(gam_1+gam_2))
    c_up_top=( source[:-1]*(1.-inv_dtaugam)+source[1:]*    inv_dtaugam )#*cst
    c_dw_top=( source[:-1]*(1.+inv_dtaugam)-source[1:]*    inv_dtaugam )#*cst
    c_up_bot=(-source[:-1]*    inv_dtaugam +source[1:]*(1.+inv_dtaugam))#*cst
    c_dw_bot=( source[:-1]*    inv_dtaugam +source[1:]*(1.-inv_dtaugam))#*cst
    #print(c_up_top, c_dw_top, c_up_bot, c_dw_bot)
    return c_up_top, c_dw_top, c_up_bot, c_dw_bot

@numba.jit(nopython=True, fastmath=True, cache=True)
def c_planck_mid(source, dtau, gam_1, gam_2, mu1):
    """c_up/dw is for c+/- whithout direct beam scattering.
    These are computed at the middle of the layer, i.e. at tau=dtau/2.
    removed a pi factor because source is pi*B
    """
    cst=1. # this factor seems to avoid emissivity greater than one. 
           # but what is the analytical basis for this ???
    inv_dtaugam=1./(dtau*(gam_1+gam_2))
    c_up_mid=cst*( source[:-1]*(.5-inv_dtaugam)+source[1:]*(0.5+inv_dtaugam))
    c_dw_mid=cst*( source[:-1]*(.5+inv_dtaugam)+source[1:]*(0.5-inv_dtaugam))
    return c_up_mid, c_dw_mid

@numba.jit(nopython=True, fastmath=True, cache=True)
def e_i_toon(dtau, gam_1, gam_2):
    """e_i factors defined in eq 44.
    """
    lamb,GAMMA=lambda_GAMMA(gam_1, gam_2)
    expdtau=np.exp(-lamb*dtau)
    #print('lamb,GAMMA,expdtau',lamb,GAMMA,expdtau)
    e_1=1.+GAMMA*expdtau
    e_2=1.-GAMMA*expdtau
    e_3=GAMMA+expdtau
    e_4=GAMMA-expdtau
    return e_1, e_2, e_3, e_4

@numba.jit(nopython=True, fastmath=True, cache=True)
def mid_factor_toon(dtau, gam_1, gam_2):
    """Factors to recover the flux at mid layer. 
    """
    lamb,GAMMA=lambda_GAMMA(gam_1, gam_2)
    expdtaumid=np.exp(-lamb*dtau*0.5)
    mid_fac1=2.*(1.+GAMMA)*expdtaumid
    mid_fac2=2.*(1.-GAMMA)*expdtaumid
    return mid_fac1, mid_fac2

@numba.jit(nopython=True, fastmath=True, cache=True)
def DTRIDGL(L,AF,BF,CF,DF):
    """
          DIMENSION AF(L),BF(L),CF(L),DF(L),XK(L)
          DIMENSION AS(2*L),DS(2*L)

    !*    THIS SUBROUTINE SOLVES A SYSTEM OF TRIDIAGIONAL MATRIX
    !*    EQUATIONS. THE FORM OF THE EQUATIONS ARE:
    !*    A(I)*X(I-1) + B(I)*X(I) + C(I)*X(I+1) = D(I)
    !======================================================================!
    """
    AS=np.empty_like(AF)
    DS=np.empty_like(AF)
    XK=np.empty_like(AF)
    AS[-1] = AF[-1]/BF[-1]
    DS[-1] = DF[-1]/BF[-1]

    for I in range(1,L):
        X         = 1./(BF[L+1-I-2] - CF[L+1-I-2]*AS[L+2-I-2])
        AS[L+1-I-2] = AF[L+1-I-2]*X
        DS[L+1-I-2] = (DF[L+1-I-2]-CF[L+1-I-2]*DS[L+2-I-2])*X
 
    XK[0]=DS[0]
    for I in range(1,L):
        XKB   = XK[I-1]
        XK[I] = DS[I]-AS[I]*XKB
    return XK
    
#@numba.jit(nopython=True,fastmath=True)
def matrix_toon(Nlay, source, dtau, gam_1, gam_2, mu1, flux_top_dw, alb_surf):
    """
    Returns
    -------
        flux_net: array
            Net flux at the bottom of the Nlay layers.
    """
    e_1, e_2, e_3, e_4 = e_i_toon(dtau, gam_1, gam_2)
    c_up_top, c_dw_top, c_up_bot, c_dw_bot = c_planck(source, dtau, gam_1, gam_2, mu1)
    A=np.empty((2*Nlay))
    B=np.empty((2*Nlay))
    D=np.empty((2*Nlay))
    E=np.empty((2*Nlay))
    # upper boundary
    # A[0]=0. #no need because of the way diagonals are treated (see solve_banded)
    B[0]=e_1[0]
    D[0]=0.
    D[1]=-e_2[0]
    E[0]=flux_top_dw-c_dw_top[0]
    #even 
    A[:-2:2]=e_1[:-1]*e_2[1:]-e_3[:-1]*e_4[1:]
    B[1:-1:2]=e_2[:-1]*e_2[1:]-e_4[:-1]*e_4[1:]
    D[2:-1:2]=e_1[1:]*e_4[1:]-e_2[1:]*e_3[1:]
    E[1:-1:2]=(c_up_top[1:]-c_up_bot[:-1])*e_2[1:]-(c_dw_top[1:]-c_dw_bot[:-1])*e_4[1:]
    # middle sign above different in my calculations and toon (+ in Toon)
    #odd
    A[1:-2:2]=e_2[:-1]*e_3[:-1]-e_4[:-1]*e_1[:-1]
    B[2:-1:2]=e_1[:-1]*e_1[1:]-e_3[:-1]*e_3[1:]
    D[3::2]=e_3[:-1]*e_4[1:]-e_1[:-1]*e_2[1:]
    E[2:-1:2]=(c_up_top[1:]-c_up_bot[:-1])*e_3[:-1]-(c_dw_top[1:]-c_dw_bot[:-1])*e_1[:-1]
    #surface
    A[-2]=e_1[-1]-alb_surf*e_3[-1]
    B[-1]=e_2[-1]-alb_surf*e_4[-1]
    #D[-1]=0.
    E[-1]=(1.-alb_surf)*source[-1]-c_up_bot[-1]+alb_surf*c_dw_bot[-1]
    #return mat,E
    Y=solve_banded((1,1),[D,B,A],E)
    flux_net = np.empty((Nlay+1))
    J4pimu   = np.empty((Nlay+1))
    flux_net[1:] = Y[::2]*(e_1-e_3)+Y[1::2]*(e_2-e_4)+c_up_bot-c_dw_bot
    J4pimu[1:]   = Y[::2]*(e_1+e_3)+Y[1::2]*(e_2+e_4)+c_up_bot+c_dw_bot
    flux_net[0]  = Y[0]*e_3[0]-Y[1]*e_4[0]+c_up_top[0]
    J4pimu[0]    = flux_net[0] + flux_top_dw
    flux_net[0] -= flux_top_dw
    return flux_net, J4pimu



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A class to handle continuum absorption (CIA)
"""

import os.path
import h5py
import numpy as np
from exo_k.util.filenames import EndOfFile
from .util.interp import linear_interpolation, interp_ind_weights, unit_convert
from .util.cst import KBOLTZ
from .settings import Settings
from .util.spectral_object import Spectral_object

class Cia_table(Spectral_object):
    """A class to handle CIA opacity data tables.

    .. important::
        In the 2018 release of Hitran, some CIA files can contain data for multiple
        wavenumber range but for different temperatures in each range.
        This is not yet supported so only the first wavenumber range is used.
        For H2-H2 cia files, one should manually replace all the `n-H2 -- n-H2` in 
        `H2-H2` for the format to be readable. 
    """

    def __init__(self, *filename_filters, filename=None, molecule_pair=None, search_path=None,
            mks=False, remove_zeros=False, old_cia_unit='cm^5'):
        """Initialization for Cia_tables.

        Parameters
        ----------
            filename: str, optional
                Relative or absolute name of the file to be loaded. 
            filename_filters: sequence of string
                As many strings as necessary to uniquely define
                a file in the global search path defined with
                :func:`exo_k.settings.Settings.set_cia_search_path`.
                This path will be searched for a file
                with all the filename_filters in the name.
                The filename_filters can contain '*'.
            molecule_pair: list of size 2, optional
                The molecule pair we want to consider, 
                specified as an array with two strings (like ['H2','H2'] or ['N2','H2O']).
                The order of the molecules in the pair is irrelevant. 

        Other Parameters
        ----------------
            old_cia_unit : str, optional
                String to specify the current cia unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the cia grid and the cia unit do not correspond).
                Available units are: 'cm^5', 'cm^2' that stand for cm^2/amagat,
                and 'cm^-1' that stand for cm^-1/amagat^2.
            remove_zeros: boolean, optional
                If True, the zeros in the kdata table are replaced by
                    a value 10 orders of magnitude smaller than the smallest positive value
            search_path: str, optional
                If search_path is provided,
                it locally overrides the global search path
                defined with :func:`exo_k.settings.Settings.set_cia_search_path`
                and only files in `search_path` are returned.
        """
        self._init_empty()
        self._settings=Settings()
        if filename is not None:
            self.filename=filename
        elif filename_filters or molecule_pair is not None:
            # a none empty sequence returns a True in a conditional statement
            self.filename=self._settings.list_cia_files(*filename_filters,
                molecule_pair=molecule_pair, only_one=True, search_path=search_path)[0]

        if self.filename is not None:
            if self.filename.lower().endswith(('h5','hdf5')):
                self.read_hdf5(self.filename)
            elif self.filename.lower().endswith('cia'):
                self.read_hitran_cia(self.filename, old_cia_unit=old_cia_unit)
            elif self.filename.lower().endswith('.dat'):
                self.read_CKD_cia(self.filename, old_cia_unit=old_cia_unit)
            else:
                raise RuntimeError('Cia file extension not known.')
        if self.abs_coeff is not None:
            if self._settings._convert_to_mks or mks: self.convert_to_mks()
            if remove_zeros : self.remove_zeros()

    def _init_empty(self):
        """Initializes attributes to none.
        """
        self.mol1=None
        self.mol2=None
        self.wns=None
        self.wnedges=None
        self.tgrid=None
        self.abs_coeff=None
        self.abs_coeff_unit='unspecified'
        self.wn_unit='cm^-1'
        self.Nt=None
        self.Nw=None
        self.filename=None


    def read_hitran_cia(self, filename, old_cia_unit='cm^5'):
        """Reads hitran cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
            old_cia_unit: str, optional
                Units found in the file.
        """
        tmp_tgrid=[]
        tmp_abs_coeff=[]
        First=True
        with open(filename, 'r') as file:
            while True:
                try:
                    Nw, Temp = self._read_header(file)
                except EndOfFile:
                    break
                except:
                    print("""
                    The .cia format your are using must not be supported.
                    This can happen with the 2018 release of the hitran CIA coefficients.
                    In particular, for some couple of molecules, the name is written with
                    spaces but exo_k only recognizes formats without spaces for now
                    (for example: \'n-H2 -- n-H2\' should be \'H2-H2\').
                    You can try replacing these in the files or contact me.
                    """)
                    raise RuntimeError('Bad .cia file format.')
                tmp_abs_coeff2 = list()
                if First: 
                    tmp_wns=[]
                    for _ in range(Nw): #as iterator not used, can be replaced by _
                        line=file.readline()
                        tmp=line.split()
                        tmp_wns.append(float(tmp[0]))
                        tmp_abs_coeff2.append(float(tmp[1]))
                    self.wns=np.array(tmp_wns)
                    self.Nw=self.wns.size
                    First = False
                else:
                    if Nw != self.Nw:
                        break # in 2018 format, there can be data for several
                              # wavenumber range in the same file
                              # but not for the same temperatures.
                              # We take only the first set for now.
                    for _ in range(Nw):
                        line=file.readline()
                        tmp=line.split()
                        tmp_abs_coeff2.append(float(tmp[1]))
                tmp_abs_coeff.append(tmp_abs_coeff2)
                tmp_tgrid.append(Temp)
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.tgrid=np.array(tmp_tgrid)
        self.abs_coeff=np.array(tmp_abs_coeff)
        self.abs_coeff_unit=old_cia_unit
        self.Nt=self.tgrid.size


    def _read_header(self, file):
        """Reads the header lines in a Hitran CIA file.

        Parameters
        ----------
            file: file stream
                file to be read.
        """
        line=file.readline()
        if line is None or line=='':
            raise EndOfFile
        tmp = line.split()
        self.mol1,self.mol2 = tmp[0].split('-')
        Nw = int(tmp[3])
        Temp = float(tmp[4])

        return Nw, Temp

    def read_hdf5(self, filename):
        """Reads hdf5 cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        f = h5py.File(filename, 'r')
        self.wns=f['bin_centers'][...]
        self.abs_coeff=f['abs_coeff'][...]
        self.abs_coeff_unit=f['abs_coeff'].attrs['units']
        self.tgrid=f['t'][...]
        if 'cia_pair' in f:
            tmp=f['cia_pair'][()]
            if isinstance(tmp, bytes): tmp=tmp.decode('UTF-8')
            self.mol1,self.mol2=tmp.split('-')
        elif 'cia_pair' in f.attrs:
            self.mol1,self.mol2=f.attrs['cia_pair'].split('-')
        f.close()  
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nt=self.tgrid.size
        self.Nw=self.wns.size
          
    def write_hdf5(self, filename):
        """Writes hdf5 cia files.

        Parameters
        ----------
            filename: str
                Name of the file to be written.
        """
        if not filename.lower().endswith(('.hdf5', '.h5')):
            filename=filename+'.h5'
        f = h5py.File(filename, 'w')
        f.create_dataset("bin_centers", data=self.wns,compression="gzip")
        f.create_dataset("t", data=self.tgrid,compression="gzip")
        f.create_dataset("abs_coeff", data=self.abs_coeff,compression="gzip")
        f["abs_coeff"].attrs["units"] = self.abs_coeff_unit
        f.create_dataset("cia_pair", data=self.mol1+'-'+self.mol2)
        f.close() 

    def write_cia(self, filename):
        """Writes simplified .cia files.

        To be consistent with Hitran CIA files, data is converted back to cm^5
        before writing.

        Parameters
        ----------
            filename: str
                Name of the file to be written.
        """
        if not filename.lower().endswith(('.cia')):
            filename=filename+'.cia'
        _,conversion_factor=unit_convert( \
            'abs_coeff_unit',unit_file=self.abs_coeff_unit,
            unit_in=self.abs_coeff_unit,unit_out='cm^5')
        to_write=conversion_factor*self.abs_coeff
        with open(filename, "w") as f:
            for i_t, t in enumerate(self.tgrid):
                header="""       {m1}-{m2}   {w1}  {w2}  {Nw}  {T}\n""".format(
                    m1=self.mol1, m2=self.mol2, w1=self.wns[0],
                    w2=self.wns[-1], Nw=self.Nw, T=t)
                f.write(header) 
                for i_w, wn in enumerate(self.wns):
                    line=str(wn)+' '+str(to_write[i_t,i_w])+'\n'
                    f.write(line) 
   

    def sample(self, wngrid, remove_zeros=False, use_grid_filter=False, **kwargs):
        """Method to re sample a cia table to a new grid of wavenumbers (inplace).

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)
            use_grid_filter: boolean, optional
                If true, the table is sampled only within the boundaries
                of its current wavenumber grid. The coefficients are set to zero elswere
                (except if remove_zeros is set to True).
                If false, the values at the boundaries are used when sampling outside the grid.
        """
        wngrid=np.array(wngrid)
        #min_val=np.amin(self.abs_coeff)
        Nnew=wngrid.size
        #wngrid_filter = np.where((wngrid <= self.wnedges[-1]) & (wngrid >= self.wnedges[0]))[0]
        if use_grid_filter:
            wngrid_filter = np.where((wngrid <= self.wns[-1]) & (wngrid >= self.wns[0]))[0]
        else:
            wngrid_filter = np.ones(Nnew,dtype=bool)
        new_abs_coeff=np.zeros((self.Nt,Nnew))
        for iT in range(self.Nt):
            tmp=self.abs_coeff[iT,:]
            new_abs_coeff[iT,wngrid_filter]=np.interp(wngrid[wngrid_filter],self.wns,tmp)
        self.abs_coeff=new_abs_coeff
        self.wns=wngrid
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.Nw=Nnew
        if remove_zeros : self.remove_zeros(**kwargs)

    def sample_cp(self, wngrid, **kwargs):
        """Creates a copy of the object before resampling it.

        Parameters
        ----------
            See sample method for details. 

        Returns
        -------
            :class:`Cia_table` object
                the re-sampled :class:`Cia_table`
        """
        res=self.copy()
        res.sample(wngrid, **kwargs)
        return res

    def interpolate_cia(self, t_array=None, log_interp=None, wngrid_limit=None):
        """interpolate_cia interpolates the kdata at on a given temperature profile. 

        Parameters
        ----------
            t_array: float or array
                Temperature array to interpolate to.
                If a float is given, it is interpreted as an array of size 1.
            wngrid_limit: array, optional
                If an array is given, interpolates only within this array.
            log_interp: bool, optional
                Whether the interpolation is linear in kdata or in log(kdata).

        Returns
        -------
            array of shape (logp_array.size, self.Nw)
                The interpolated kdata.
        """
        if hasattr(t_array, "__len__"):
            t_array=np.array(t_array)
        else:
            t_array=np.array([t_array])
        tind,tweight=interp_ind_weights(t_array,self.tgrid)
        if wngrid_limit is None:
            wngrid_filter = slice(None)
            Nw=self.Nw
        else:
            wngrid_limit=np.array(wngrid_limit)
            wngrid_filter = np.where((self.wnedges > wngrid_limit.min()) & (
                self.wnedges <= wngrid_limit.max()))[0][:-1]
            Nw=wngrid_filter.size
        res=np.zeros((tind.size,Nw))
        if log_interp is None: log_interp=self._settings._log_interp
        if log_interp:
            for ii in range(tind.size):
                kc_t1=np.log(self.abs_coeff[tind[ii]][wngrid_filter].ravel())
                kc_t0=np.log(self.abs_coeff[tind[ii]-1][wngrid_filter].ravel())
                res[ii]=linear_interpolation(kc_t0, kc_t1, tweight[ii])
            return np.exp(res)
        else:
            for ii in range(tind.size):
                #kc_t1=self.abs_coeff[tind[ii]]
                kc_t1=self.abs_coeff[tind[ii]][wngrid_filter].ravel()
                kc_t0=self.abs_coeff[tind[ii]-1][wngrid_filter].ravel()
                res[ii]=linear_interpolation(kc_t0, kc_t1, tweight[ii])
            return res

    def equivalent_xsec(self, logP, T, x_mol2, wngrid_limit=None):
        """Computes the cross section due to CIA in area per molecule of type 1.
        """
        n_density=10**logP/(KBOLTZ*T)
        return self.interpolate_cia(t_array=T,wngrid_limit=wngrid_limit)*n_density*x_mol2

    def effective_cross_section(self, logP, T, x_mol1, x_mol2, wngrid_limit=None):
        """Computes the total cross section for a molecule pair
        (in m^2 per total number of molecules; assumes data in MKS).

        Parameters
        ----------
            logP: float or array
                Log10 of the pressure (Pa).
            T: float or array
                Temperature (K).
            x_mol1/2: float or array
                Volume mixing ratio of the 1st and 2nd molecule of the pair.
            wngrid_limit: array, optional  
                If an array is given, interpolates only within this array. 

        Returns
        -------
            float or array
                total cross section for the molecule pair in m^2 per total number of molecules.
        """
        x_x_n_density=10**logP/(KBOLTZ*T)*x_mol1*x_mol2
        #return self.interpolate_cia( \
        # T,wngrid_limit=wngrid_limit)*n_density[:,None]*n_density[:,None]*x_mol1*x_mol2
        tmp=self.interpolate_cia(t_array=T, wngrid_limit=wngrid_limit)
        return (x_x_n_density*tmp.transpose()).transpose() 
        # trick for the broadcasting to work whether x_x_n_density is a float or an array

    def plot_spectrum(self, ax, t=200., x_axis='wls', xscale=None, yscale=None, **kwarg):
        """Plot the spectrum for a given point

        Parameters
        ----------
            ax : :class:`pyplot.Axes`
                A pyplot axes instance where to put the plot.
            t: float
                temperature(K)
            x_axis: str, optional
                If 'wls', x axis is wavelength. Wavenumber otherwise.
            x/yscale: str, optional
                If 'log' log axes are used.
        """
        toplot=self.interpolate_cia(t)[0]
        if x_axis == 'wls':
            ax.plot(self.wls,toplot,**kwarg)
            ax.set_xlabel('Wavelength (micron)')
        else:
            ax.plot(self.wns,toplot,**kwarg)
            ax.set_xlabel('Wavenumber (cm$^{-1}$)')
        ax.set_ylabel('Abs. coeff')
        if xscale is not None: ax.set_xscale(xscale)
        if yscale is not None: ax.set_yscale(yscale)

    def convert_abs_coeff_unit(self,abs_coeff_unit='unspecified',old_abs_coeff_unit='unspecified'):
        """Converts abs_coeff to a new unit (inplace).

        Parameters
        ----------
            abs_coeff_unit: str
                String to identify the units to convert to.
                Accepts 'cm^5', 'm^5'. or any length^5 unit recognized by the 
                astropy.units library. If ='unspecified', no conversion is done.
            old_abs_coeff_unit : str, optional
                String to specify the current kdata unit if it is unspecified or if 
                you have reasons to believe it is wrong (e.g. you just read a file where
                you know that the kdata grid and the kdata unit do not correspond)
        """
        if abs_coeff_unit==old_abs_coeff_unit: return
        tmp_k_u_in=old_abs_coeff_unit
        tmp_k_u_out=abs_coeff_unit
        tmp_k_u_file=self.abs_coeff_unit
        self.abs_coeff_unit,conversion_factor=unit_convert( \
            'abs_coeff_unit',unit_file=tmp_k_u_file,unit_in=tmp_k_u_in,unit_out=tmp_k_u_out)
        self.abs_coeff=self.abs_coeff*conversion_factor

    def convert_to_mks(self, verbose=False):
        """Converts units to MKS
        """
        message=None
        if self.abs_coeff_unit=='cm^5':
            message='Conversion from cm^5 to m^5'
            self.convert_abs_coeff_unit(abs_coeff_unit='m^5')
        elif self.abs_coeff_unit in ['cm^2','m^2']:
            message='Conversion from '+self.abs_coeff_unit+'/amagat to m^5'
            self.convert_abs_coeff_unit(abs_coeff_unit='m^2')
            self.abs_coeff=self.abs_coeff*(KBOLTZ*273.15/101325.0)
            #conversion from m^2/amagat to m^5
        elif self.abs_coeff_unit in ['cm^-1','m^-1']:
            message='Conversion from '+self.abs_coeff_unit+'/amagat^2 to m^5'
            self.convert_abs_coeff_unit(abs_coeff_unit='m^-1')
            self.abs_coeff=self.abs_coeff*(KBOLTZ*273.15/101325.0)**2
            #conversion from m^-1/amagat^2 to m^5
        else:
            return
        if verbose and message is not None:
            print(message)
        self.abs_coeff_unit='m^5'
        return

    def remove_zeros(self, deltalog_min_value=0.):
        """Finds zeros in the abs_coeff and set them to (10.^-deltalog_min_value)
        times the minimum positive value in the table (inplace).
        This is to be able to work in logspace. 
        """
        mask = np.zeros(self.abs_coeff.shape,dtype=bool)
        mask[np.nonzero(self.abs_coeff)] = True
        minvalue=np.amin(self.abs_coeff[mask])
        self.abs_coeff[~mask]=minvalue/(10.**deltalog_min_value)
        
    def __repr__(self):
        """Method to output header
        """
        output="""
        file          : {file}
        molecule pair : {mol1} - {mol2}
        t grid   (K)  : {t}
        abs coeff     : {abs_coeff}
        """.format(file=self.filename,mol1=self.mol1, mol2=self.mol2, \
            t=self.tgrid,abs_coeff=self.abs_coeff)
        return output

    def __getitem__(self,key):
        """Overrides getitem.
        """
        return self.abs_coeff[key]

    def copy(self):
        """Creates a new instance of :class:`CIA_table` object and (deep) copies data into it
        """
        res=Cia_table()
        res.mol1 = self.mol1
        res.mol2 = self.mol2
        res.wns = np.copy(self.wns)
        res.wnedges = np.copy(self.wnedges)
        res.tgrid = np.copy(self.tgrid)
        res.abs_coeff = np.copy(self.abs_coeff)
        res.abs_coeff_unit = self.abs_coeff_unit
        res.Nt = self.Nt
        res.Nw = self.Nw
        res.filename = self.filename
        res._settings=Settings()
        return res

    def read_CKD_cia(self, filename, old_cia_unit='cm^2'):
        """Reads hitran cia files and load temperature, wavenumber, and absorption coefficient grid.

        Parameters
        ----------
            filename: str
                Name of the file to be read.
        """
        self.tgrid=np.array([200., 250., 300., 350., 400.,
            450., 500., 550., 600., 650., 700.])
        self.Nt=self.tgrid.size
        self.abs_coeff=np.loadtxt(filename, skiprows=1, unpack=True)
        self.mol1='H2O'
        if 'SELF' in os.path.basename(filename):
            self.mol2='H2O'
        else:
            self.mol2='N2'
        nu_name=os.path.join(os.path.dirname(filename),'H2O_CONT_NU.dat')
        self.wns=np.loadtxt(nu_name, skiprows=1, unpack=True)
        self.wnedges=np.concatenate(([self.wns[0]],0.5*(self.wns[1:]+self.wns[:-1]),[self.wns[-1]]))
        self.abs_coeff_unit=old_cia_unit
        self.Nw=self.wns.size

#    #  amagatS=(273.15/temp)*(presS/101325.0)
#    #  amagatF=(273.15/temp)*(presF/101325.0)
#
#    #  abcoef = abcoefS*amagatS + abcoefF*amagatF ! Eq. (15) in Clough (1989)
#    #  abcoef = abcoef*(presS/(presF+presS))      ! take H2O mixing ratio into account
#    #                                                ! abs coeffs are given per molecule of H2O
#
#    #  Nmolec = (presS+presF)/(kB*temp)           ! assume ideal gas
#!      print*,'Total number of molecules per m^3 is',Nmolec
#
#      abcoef = (abcoefS*x_H2O + abcoefF*(1-x_other))*(k_b*273.15/101325.0)*x_H2O
#                     *Nmolec*2/(100.0**2)          ! convert to m^-1

    def effective_cross_section2(self, logP, T, x_mol1, x_mol2, wngrid_limit=None):
        """Obsolete.
        
        Computes the total cross section for a molecule pair
        (in m^2 per total number of molecules; assumes data in MKS).
        """
        x_x_n_density=10**logP/(KBOLTZ*T)*x_mol1*x_mol2
        #return self.interpolate_cia( \
        # T,wngrid_limit=wngrid_limit)*n_density[:,None]*n_density[:,None]*x_mol1*x_mol2
        tmp=self.interpolate_cia(t_array=T, wngrid_limit=wngrid_limit)
        return x_x_n_density[:,None]*tmp

